// ChartContainer.cpp : implementation file
//

#include "stdafx.h"
#include <ctime>
#include "ChartDef.h"
#include "Chart.h"
#include "DataLabel.h"
#include "ChartContainer.h"
#include "ChartDataView.h"
#include "ChartsXMLSerializer.h"
#include "Util.h"
#include <Shlwapi.h>

using namespace std;
using namespace Gdiplus;

///////////////////////////////////////////////////////////////////////////////
// CChartContainer

IMPLEMENT_DYNAMIC(CChartContainer, CWnd)

CChartContainer::CChartContainer(string_t name) : m_colBkground((ARGB)Color::White), 
                                     m_colBorder((ARGB)Color::Black),
                                     m_colAxis((ARGB)Color::Black), 
                                     m_colZoomBorder((ARGB)Color::Red), 
                                     m_colLegPnts((ARGB)Color::Red),
                                     m_colDataViewPnts((ARGB)Color::BlueViolet), 
                                     m_colLegBkgnd(Color(0xFFFFFFE0)),
                                     m_colLegBorder((ARGB)Color::Black),
                                     m_colGrid((ARGB)Color::Gray), 
                                     m_minExtY(DBL_MAX), m_maxExtY(-DBL_MAX), 
                                     m_startX(DBL_MAX),m_endX(-DBL_MAX), 
                                     m_axOffsXF(0.0f), m_axXPos(BETWEEN), m_axYPos(MIDDLE), 
                                     m_scX(1.0), m_scY(1.0), 
                                     m_bTracking(false), m_bSelEnabled(false),  m_bAxBoundaries(false),
                                     m_chartSelIdx(-1), m_dataViewChartIdx(-1), m_precision(3),
                                     m_tmpZX(0.0), 
                                     m_pMatrixD(NULL), 
                                     m_dataLegPntD(PointD(DBL_MAX, DBL_MAX)), m_epsX(0.0), 
                                     m_pDataWnd(NULL), m_pLegWnd(NULL),
                                     m_pChartDataView(NULL), m_chMode(MODE_FULLX), m_name(name), 
                                     m_labelX(string_t(_T("X"))), m_pLabValStrFnPtr(&GetLabelValStr)
{
}

CChartContainer::~CChartContainer()
{
  if (m_pChartDataView != NULL)
  {
    delete m_pChartDataView;
    m_pChartDataView = NULL;
  }

  DestroyLabWnd(m_pDataWnd);
  DestroyLabWnd(m_pLegWnd);

  if (m_pMatrixD != NULL)
    delete m_pMatrixD;

  if (!m_mapCharts.empty())
  {
    MAP_CHARTS::iterator it = m_mapCharts.begin();
    MAP_CHARTS::iterator itE = m_mapCharts.end();
    for (; it != itE; ++it)
    {
      delete it->second;
      it->second = NULL;
    }
  }
}

CChartContainer& CChartContainer::operator=(const CChartContainer& rhs)
{
  if (this != &rhs)
  {
    m_pDataWnd = DestroyLabWnd(m_pDataWnd); // NULL
    m_pLegWnd  = DestroyLabWnd(m_pLegWnd);  //NULL

    m_colBkground   = rhs.m_colBkground;
    m_colBorder     = rhs.m_colBorder;
    m_colAxis       = rhs.m_colAxis;
    m_colZoomBorder = rhs.m_colZoomBorder;
    m_colLegPnts    = rhs.m_colLegPnts;
    m_colDataViewPnts = rhs.m_colDataViewPnts;
    m_colLegBkgnd   = rhs.m_colLegBkgnd;
    m_colLegBorder  = rhs.m_colLegBorder;
    m_colGrid       = rhs.m_colGrid;

    m_minExtY       = rhs.m_minExtY;
    m_maxExtY       = rhs.m_maxExtY;
    m_startX        = rhs.m_startX;
    m_endX          = rhs.m_endX;

    m_axOffsXF      = 0.0f;         // Can't be defined until container wnd is created
    m_axXPos        = BETWEEN;
    m_axYPos        = MIDDLE;

    m_scX = 1.0;     // Will be calculated in OnPaint()
    m_scY = 1.0; 
   
    m_bTracking       = rhs.m_bTracking;    // Call ShowLegend after wnd creation to see
    m_dataLegPntD     = rhs.m_dataLegPntD;  // data legend window
    m_epsX            = rhs.m_epsX;         // Neighbourhood to search for nearest pnts
    m_bSelEnabled     = false;
    m_bAxBoundaries   = rhs.m_bAxBoundaries, 
    m_chartSelIdx     = rhs.m_chartSelIdx; 
    m_dataViewChartIdx = -1;  // Data View 

    m_pMatrixD        = NULL; // To calculate in OnPaint()
   
    m_precision       = rhs.m_precision;
    m_chMode          = MODE_FULLX;

    m_name            = rhs.m_name;
    m_labelX          = rhs.m_labelX.empty() ? string_t(_T("X")) : rhs.m_labelX;;
    m_pLabValStrFnPtr = rhs.m_pLabValStrFnPtr;

// Copy charts 
    m_mapCharts.clear();
    MAP_CHARTS::const_iterator itCharts = rhs.m_mapCharts.begin();
    MAP_CHARTS::const_iterator itChartsE = rhs.m_mapCharts.end();
    for (; itCharts != itChartsE; ++itCharts)
    {
      CChart* chartPtr = itCharts->second->CloneChart(true);
      m_mapCharts.insert(MAP_CHARTS::value_type(itCharts->first, chartPtr));
    }

    m_vHist = rhs.m_vHist;
    m_mapSelPntsD = rhs.m_mapSelPntsD;
    m_mapDataViewPntsD.clear();

    if (IsWindow(m_hWnd) && IsWindowVisible())
    {
      if (rhs.IsLabWndVisible(false))
        CopyDataLegendWnd();
      else
        RefreshWnd();

      if (rhs.IsLabWndVisible(true))
        ShowNamesLegend();
    }
  }
  return *this;
}

bool CChartContainer::CreateChartCtrlWnd(DWORD dwExStyle, DWORD dwStyle, 
                                        const CRect& wndRect, CWnd* pParent, UINT nID)
{
     BOOL bRes = CWnd::CreateEx(dwExStyle,
                                AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_SAVEBITS),
                                NULL,
                                WS_CHILD|WS_CLIPCHILDREN|WS_VISIBLE|dwStyle,
                                wndRect,
                                pParent,
                                nID
                               );
     return bRes == TRUE ? true : false;
  
 }

///////////////////////////////////////////////////////////////////////////////
// Helpers and operations

CChartContainer* CChartContainer::CloneChartContainer(string_t name, bool bCloneState)
{
  CChartContainer* pContainer = new CChartContainer;

  if (name.empty())
    pContainer->m_name   = m_name;
  else
    pContainer->m_name   = name;

  MAP_CHARTS::iterator itCharts = m_mapCharts.begin();
  MAP_CHARTS::iterator itChartsE = m_mapCharts.end();
  for (; itCharts != itChartsE; ++itCharts)
  {
    CChart* cloneChartPtr = itCharts->second->CloneChart(bCloneState);
    pContainer->m_mapCharts.insert(MAP_CHARTS::value_type(cloneChartPtr->GetChartIdx(),
                                                                        cloneChartPtr));
  }

  pContainer->m_colBkground   = m_colBkground;
  pContainer->m_colBorder     = m_colBorder;
  pContainer->m_colAxis       = m_colAxis;
  pContainer->m_colZoomBorder = m_colZoomBorder;
  pContainer->m_colLegPnts    = m_colLegPnts;
  pContainer->m_colDataViewPnts = m_colDataViewPnts;
  pContainer->m_colLegBkgnd   = m_colLegBkgnd;
  pContainer->m_colLegBorder  = m_colLegBorder;
  pContainer->m_colGrid       = m_colGrid;

  pContainer->m_minExtY       = m_minExtY;
  pContainer->m_maxExtY       = m_maxExtY;

  pContainer->m_bTracking     = m_bTracking;
  pContainer->m_dataLegPntD   = m_dataLegPntD;
  pContainer->m_epsX          = m_epsX;

  pContainer->m_labelX         = m_labelX;
  pContainer->m_pLabValStrFnPtr   = m_pLabValStrFnPtr;

 if (bCloneState)
  {
    pContainer->m_startX        = m_startX;
    pContainer->m_endX          = m_endX;
    pContainer->m_vHist        = m_vHist;
    pContainer->m_bSelEnabled   = m_bSelEnabled;
    pContainer->m_chartSelIdx    = m_chartSelIdx;
    pContainer->m_bAxBoundaries = m_bAxBoundaries;

    if (m_bTracking && IsLabWndExist(false))
      pContainer->m_mapSelPntsD = m_mapSelPntsD;
      
  }
  else
  {
    pContainer->m_startX        = GetInitialStartX();
    pContainer->m_endX          = GetInitialEndX();
  }

  if (!pContainer->m_vHist.empty())
    pContainer->m_chMode = MODE_ZOOMED;

  pContainer->m_precision     = m_precision;

  return pContainer;
}

CChart* CChartContainer::CloneChart(int chartIdx, bool bCloneState)
{
  MAP_CHARTS::iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.end())
    return it->second->CloneChart(bCloneState);
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
// Setting colors

void CChartContainer::SetBkColor(Color col, bool bRedraw)
{
  m_colBkground = col;
  if (bRedraw)
    RefreshWnd();
}

void CChartContainer::SetBorderColor(Color col, bool bRedraw)
{
  m_colBorder = col;
  if (bRedraw)
    RefreshWnd();
}

void CChartContainer::SetAxesColor(Color col, bool bRedraw)
{
  m_colAxis = col;
  if (bRedraw)
    RefreshWnd();
}

void CChartContainer::SetZoomBorderColor(Color col)
{
  if (m_colZoomBorder.GetValue() != col.GetValue())
  {
    m_colZoomBorder = col;
    if (m_chMode == MODE_ZOOMING)
    {
      PointD tmpPntD(m_tmpZX, 0.0);      
      PointF tmpPntF = m_pMatrixD->TransformToPntF(1.0, tmpPntD);
      CPoint tmpPnt  = CPointFromPntF(tmpPntF);
      DrawZoomBorder(tmpPnt.x);
    }
  }
}

void CChartContainer::SetLegPntsColor(Color col, bool bRedraw)
{
  m_colLegPnts = col;
  if (bRedraw && IsLabWndVisible(false))
    RefreshWnd();
}

void CChartContainer::SetDataViewPntsColor(Color col, bool bRedraw)
{
  m_colDataViewPnts = col;
  if (bRedraw && (!m_mapDataViewPntsD.empty()))
    RefreshWnd();
}

void CChartContainer::SetLegBkgndColor(Color col, bool bRedraw)
{
  m_colLegBkgnd = col;
  if (m_pLegWnd != NULL)
  {
    m_pLegWnd->SetBkColor(col);
    if (bRedraw && m_pLegWnd->IsWindowVisible())
          RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE);
  }

  if (m_pDataWnd != NULL)
  {
    m_pDataWnd->SetBkColor(col);
    if (bRedraw && m_pDataWnd->IsWindowVisible())
          RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE);
  }
}

void CChartContainer::SetLegBorderColor(Color col, bool bRedraw)
{
  m_colLegBorder = col;
  if (m_pLegWnd != NULL)
  {
    m_pLegWnd->SetBorderColor(col);
    if (bRedraw && m_pLegWnd->IsWindowVisible())
          RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE);
  }

  if (m_pDataWnd != NULL)
  {
    m_pDataWnd->SetBorderColor(col);
    if (bRedraw && m_pDataWnd->IsWindowVisible())
          RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE);
  }
}

void CChartContainer::SetGridColor(Color col, bool bRedraw)
{
  m_colGrid = col;
  if (bRedraw)
    RefreshWnd();
}

double CChartContainer::GetInitialStartX(void) const
{
  return m_vHist.empty() ? m_startX : m_vHist.front().first;
}

double CChartContainer::GetInitialEndX(void) const
{
  return m_vHist.empty() ? m_endX : m_vHist.front().second;
}

void CChartContainer::UpdateExtX(double minExtX, double maxExtX, bool bRedraw)
{
  if (maxExtX <= minExtX)                   // Possible if from app
    return;

  double initStartX = GetInitialStartX();   // Old initial m_startX, m_endX
  double initEndX   = GetInitialEndX();

  if (initStartX > initEndX)  // The container is empty
  {
    m_startX = minExtX;
    m_endX   = maxExtX;
  }
  else
  {
    double startX = min(minExtX, initStartX);
    double endX   = max(maxExtX, initEndX);

    if (m_vHist.size() > 0)   // Was zoomed or panned
      m_vHist.front() = make_pair(startX, endX);
    else                      // Has no history
    {
      m_startX = startX;
      m_endX   = endX;
    }
  }

  if (bRedraw)
  {
    if (m_bTracking&& IsLabWndExist(true))
      UpdateDataLegend(false);
    else
      RefreshWnd();
  }
}

void CChartContainer::UpdateExtY(double minExtY, double maxExtY, bool bRedraw)
{
  if (m_minExtY > minExtY)
    m_minExtY = minExtY;
  if (m_maxExtY < maxExtY)
    m_maxExtY = maxExtY;

  if (bRedraw)
  {
    if (m_bTracking&& IsLabWndExist(false))
      UpdateDataLegend(true);
    else
      RefreshWnd();
  }
}

void CChartContainer::SetContainerPrecision(int precision, bool bRedraw)
{
  m_precision = precision;
  if (bRedraw)
  {
    if (IsLabWndVisible(false)) // This version of Update redraws dataWnd only
      UpdateDataLegend(false);  //(m_mapSelPntsD, m_pDataWnd->m_mapLabs);
    else if (m_bAxBoundaries)
      RefreshWnd();
    if (m_dataViewChartIdx != -1)
      UpdateDataView(m_dataViewChartIdx, F_NODATACHANGE);
  }
}

size_t CChartContainer::SetDataViewSelPnts(int chartIdx, 
                            const MAP_SELPNTSD& mapSelPntsD, V_CHARTDATAD* pChartData)
{
  if (chartIdx == -1)
    return 0;
  if (chartIdx != m_dataViewChartIdx)
    return 0;
  if (m_pChartDataView != NULL)
  {
    m_pChartDataView->m_mapSelCells = mapSelPntsD;
    if (pChartData == NULL)
    {
      m_pChartDataView->RefreshSelCells();
      if (IsWindow(m_pChartDataView->m_hWnd))
      {
        if (m_pChartDataView->IsIconic())
          m_pChartDataView->ShowWindow(SW_RESTORE);
        m_pChartDataView->RedrawWindow(NULL, NULL, 
                RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE|RDW_ALLCHILDREN);
      }
    }
    else
      ReplaceChartData(chartIdx, *pChartData, false, false, true);
  }
  return mapSelPntsD.size();
}

bool CChartContainer::ClearDataViewChartIdx(void)
{
  if ((m_pChartDataView != NULL) &&(m_dataViewChartIdx != -1))
  {
    m_dataViewChartIdx = -1;
    UpdateDataViewPnts(-1, 0, PointD(0.0, 0.0), false);
    if (IsWindow(m_pChartDataView->m_hWnd))
      m_pChartDataView->DestroyWindow();
    delete m_pChartDataView;
    m_pChartDataView = NULL;
    return true;
  }
  return false;
}

void CChartContainer::SetAxisXName(string_t nameX, bool bRedraw)
{
  m_labelX = NormalizeString(nameX, STR_MAXLEN, STR_NORMSIGN);

  if (bRedraw)
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);
    else if (m_bAxBoundaries)
      RefreshWnd();
    if (m_dataViewChartIdx != -1)
      UpdateDataView(m_dataViewChartIdx, F_NODATACHANGE);
  }
}

void CChartContainer::SetLabXValStrFn(val_label_str_fn pLabValStrFn, bool bRedraw)
{
  m_pLabValStrFnPtr = pLabValStrFn;

  if (bRedraw)
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);
    else if (m_bAxBoundaries)
      RefreshWnd();
    if (m_dataViewChartIdx != -1)
      UpdateDataView(m_dataViewChartIdx, F_NODATACHANGE);
  }
}

MatrixD* CChartContainer::GetTransformMatrixD(double startX, double endX, double minY, double maxY)
{
  CDC* pDC = GetDC();
  Graphics gr(pDC->m_hDC);

  Rect rGdi;
  gr.GetVisibleClipBounds(&rGdi);                 // The same as the clip rect

  RectF chartRF = GdiRectToRectF(rGdi);           // To be exactly as in OnPaint
  chartRF.Width -= 1.0f;                          
  chartRF.Height -= 1.0f;
  chartRF.Inflate(-DR_SPACE, -DR_SPACE);
// Axis X: Its Y Position and Offset; also Y coord of orig point
  PAIR_XAXPOS axXPosY = GetXAxisPos(chartRF, minY, maxY);
// Axis Y: Its X Position and ofset; also X-coord of orig. point
  PAIR_YAXPOS axYPosX = GetYAxisPos(chartRF, startX, endX); 

  std::pair<double, double> pair_scales = UpdateScales(chartRF, 
                                                  startX, endX, minY, maxY);
  double scX = pair_scales.first;
  double scY = pair_scales.second;

  MatrixD matrixD;
  matrixD.Translate(axYPosX.second, axXPosY.second); 
  matrixD.Scale(float(scX), float(-scY));

  if ((axYPosX.first != BETWEEN)||(axXPosY.first != MIDDLE))
  {
    double translateX = (axYPosX.first == RIGHT) ? -endX : 
                                  (axYPosX.first == LEFT) ? -startX : 0.0;
    double translateY = (axXPosY.first == BOTTOM) ? -minY : 
                                      (axXPosY.first == TOP) ? -maxY : 0.0;
    matrixD.Translate(float(translateX), float(translateY)); 
  }

  ReleaseDC(pDC);
  return matrixD.Clone();
}

void CChartContainer::SetTransformMatrixD(MatrixD* pMatrixD) 
{ 
  if (m_pMatrixD != NULL) 
    delete m_pMatrixD; 
  m_pMatrixD = pMatrixD->Clone();
}

void CChartContainer::SetTransformMatrixD(void) 
{ 
  if (m_pMatrixD != NULL) 
    delete m_pMatrixD;
  m_pMatrixD = 
        GetTransformMatrixD(m_startX, m_endX, m_minExtY, m_maxExtY);
}

PAIR_DBLS CChartContainer::GetMinMaxExtX(void) const
{
  double fMinExtX = DBL_MAX;
  double fMaxExtX = -DBL_MAX;

  double currMinX, currMaxX;
  CChart* chartPtr;

  MAP_CHARTS::const_iterator it = m_mapCharts.cbegin();
  MAP_CHARTS::const_iterator itE = m_mapCharts.cend();
  for (; it != itE; ++it)
  {
    chartPtr = it->second;
    currMinX = chartPtr->GetMinValX();
    currMaxX = chartPtr->GetMaxValX();
    if (currMinX < fMinExtX)
      fMinExtX = currMinX;
    if (currMaxX > fMaxExtX)
      fMaxExtX = currMaxX;
  }

  return make_pair(fMinExtX, fMaxExtX);
}

PAIR_DBLS CChartContainer::GetMinMaxExtY(void) const
{
  double fMinExtY = DBL_MAX;
  double fMaxExtY = -DBL_MAX;

  double currMinY, currMaxY;
  CChart* chartPtr;

  MAP_CHARTS::const_iterator it = m_mapCharts.cbegin();
  MAP_CHARTS::const_iterator itE = m_mapCharts.cend();
  for (; it != itE; ++it)
  {
    chartPtr = it->second;
    currMinY = chartPtr->GetMinValY();
    currMaxY = chartPtr->GetMaxValY();
    if (currMinY < fMinExtY)
      fMinExtY = currMinY;
    if (currMaxY > fMaxExtY)
      fMaxExtY = currMaxY;
  }

  return make_pair(fMinExtY, fMaxExtY);
}

void CChartContainer::UndoHistStep(bool bRedraw)
{
  if (m_vHist.empty())
    return;

  PAIR_POS zh = m_vHist.back();
  m_startX = zh.first;
  m_endX = zh.second;
  m_vHist.pop_back();

// Restore state
  if (m_vHist.empty())
    m_chMode = MODE_FULLX;
  else 
    m_chMode = MODE_ZOOMED;

  if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
  {
    if (m_bTracking && (m_pDataWnd != NULL))
      UpdateDataLegend(true);
    else
      RefreshWnd();
  }
}

void CChartContainer::OnChangedSize(int cx, int cy)
{
  HideLabWnds();
  MoveWindow(0, 0, cx, cy);
  RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE);
  SetTimer(2, 50, NULL);
}

void CChartContainer::RefreshWnd(void)
{
  CRect clR, legR, dataR;
  GetClientRect(&clR);

  CRgn rgn1, rgn2, rgn3;
  rgn1.CreateRectRgn(0, 0, clR.right, clR.bottom);
  if (IsLabWndVisible(true))
  {
    m_pLegWnd->GetWindowRect(&legR);
    ScreenToClient(&legR);
    rgn2.CreateRectRgn(legR.left, legR.top, legR.right, legR.bottom);
    rgn1.CombineRgn(&rgn1, &rgn2, RGN_DIFF);
  }
  if (IsLabWndVisible(false))
  {
    m_pDataWnd->GetWindowRect(&dataR);
    ScreenToClient(&dataR);
    rgn3.CreateRectRgn(dataR.left, dataR.top, dataR.right, dataR.bottom);
    rgn1.CombineRgn(&rgn1, &rgn3, RGN_DIFF);
  }

  RedrawWindow(NULL, &rgn1, RDW_INVALIDATE | RDW_UPDATENOW | RDW_NOERASE);//|RDW_ALLCHILDREN);
}

HRESULT CChartContainer::SaveChartData(const string_t pathName, const CChartContainer* pContainer, 
                                                                            const string_t chartName)
{
  return CChartsXMLSerializer::ChartDataToXML(pathName.c_str(), pContainer, chartName);
}

HRESULT CChartContainer::SaveChartData(void)
{
  HRESULT hr = S_FALSE;
// See what you are going to save
  if (NULL == GetVisibleChart(m_chartSelIdx))
    return hr;

// Select file
  TCHAR szFilters[]= _T("Xml Files (*.xml)|*.xml");
  TCHAR szWorkDirPath[255];
  GetModuleFileName(NULL, szWorkDirPath, 255);
  PathRemoveFileSpec(szWorkDirPath);

  string_t dirStr(szWorkDirPath);
  size_t lastSlash = dirStr.find_last_of(_T("\\")) + 1;
  string_t chartDir(_T("Charts"));
  dirStr.replace(lastSlash, chartDir.length(), chartDir.c_str());

  CFileDialog fileDlg(FALSE, _T("xml"), _T("*.xml"), 
        OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT|OFN_NOCHANGEDIR|OFN_EXPLORER,
        szFilters, this);
  fileDlg.m_ofn.lpstrInitialDir = dirStr.c_str();
  
  string_t strTitle(_T("Save "));

  if (fileDlg.DoModal() == IDOK)
  {
     CString pathName = fileDlg.GetPathName();
     CString ext = fileDlg.GetFileExt();
     if (ext == CString(_T("xml")))
     {
       string_t strPath(pathName);
       CChartContainer* pContainer = CloneChartContainer(string_t(_T("")), true);
       string_t chartName;
       if (m_chartSelIdx != -1)
         chartName = pContainer->GetChartName(m_chartSelIdx);
       hr = SaveChartData(strPath, pContainer, chartName);
       delete pContainer;
     }
  }
  else
    hr = S_OK;

  return hr;
}

HRESULT CChartContainer::LoadCharts(LPCTSTR fileName, const MAP_CHARTCOLS& mapContent)
{
  HRESULT hr = CChartsXMLSerializer::XMLToCharts(fileName, this, mapContent);
  if (hr == S_OK)
  {
    if (IsLabWndExist(false))
      PrepareDataLegend(m_dataLegPntD, m_epsX, m_pDataWnd->m_mapLabs, m_mapSelPntsD, NULL);
 
    UpdateContainerWnds();
  }
  return hr;
}

HRESULT CChartContainer::GetChartNamesFromXMLFile(LPCTSTR fileName, MAP_CHARTCOLS& mapContent)
{
  return CChartsXMLSerializer::GetChartNamesFromXMLFile(fileName, mapContent);
}

HRESULT CChartContainer::GetChartNamesFromXMLFile(LPCTSTR fileName, MAP_NAMES& mapNames)
{
  return CChartsXMLSerializer::GetChartNamesFromXMLFile(fileName, mapNames);
}

bool CChartContainer::ExportChartData(string_t chartName, V_CHARTDATAD& vDataPnts) const
{
  CChart* chartPtr = FindChartByName(chartName);
  if (chartPtr != NULL)
  {
    size_t dataSize = chartPtr->m_vDataPnts.size();
    if (dataSize != 0)
    {
      vDataPnts.resize(dataSize);
      copy(chartPtr->m_vDataPnts.cbegin(), chartPtr->m_vDataPnts.cend(), vDataPnts.begin());
      return true;
    }
  }
  return false;
}

bool CChartContainer::ExportChartData(string_t chartName, 
                                std::vector<std::pair<double, double> >& vPairs) const
{
  CChart* chartPtr = FindChartByName(chartName);
  if (chartPtr != NULL)
  {
    size_t dataSize = chartPtr->m_vDataPnts.size();
    if (dataSize != 0)
    {
      vPairs.resize(dataSize);
      for (size_t i = 0; i < dataSize; ++i)
      {
        PointD pntD = chartPtr->m_vDataPnts[i];
        vPairs[i] = make_pair(pntD.X, pntD.Y);
      }
      return true;
    }
  }
  return false;
}

bool CChartContainer::ExportChartData(string_t chartName, 
                                   std::vector<double>& vX, std::vector<double>& vY) const
{
  CChart* chartPtr = FindChartByName(chartName);
  if (chartPtr != NULL)
  {
    size_t dataSize = chartPtr->m_vDataPnts.size();
    if (dataSize != 0)
    {
      vX.resize(dataSize);
      vY.resize(dataSize);
      for (size_t i = 0; i < dataSize; ++i)
      {
        PointD pntD = chartPtr->m_vDataPnts[i];
        vX[i] = pntD.X;
        vY[i] = pntD.Y;
      }
      return true;
    }
  }
  return false;
}

// Chart access and manipulation functions
/// 
/// Returns false in the third member of the tuple if the chart does not exist
///
std::tuple<double, double, bool> CChartContainer::GetChartBoundaries(int chartIdx, bool bY) const
{
  MAP_CHARTS::const_iterator itCh = m_mapCharts.find(chartIdx);
  if (itCh != m_mapCharts.cend())
    return bY == false ? make_tuple(itCh->second->GetMinValX(),itCh->second->GetMaxValX(), true):
                         make_tuple(itCh->second->GetMinValY(),itCh->second->GetMaxValY(), true);
  return make_tuple(0.0, 0.0, false);
}

CChart* CChartContainer::GetVisibleChart(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = (chartIdx) == -1 ? m_mapCharts.cbegin() :
                                                m_mapCharts.find(chartIdx);
  for (; it != m_mapCharts.cend(); ++it)
  {
    CChart* chartPtr = it->second;
    if (chartPtr->IsChartVisible())
      return chartPtr;
    if (chartIdx != -1)
      break;
  }
  return NULL;
}

bool CChartContainer::IsChartVisible(int chartIdx) const
{
  return GetVisibleChart(chartIdx) == NULL ? false : true;
}

bool CChartContainer::SetChartVisibility(int chartIdx, bool bVisible, bool bRedraw)
{
  MAP_CHARTS::iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.end())
  {
    bool bVisState = it->second->IsChartVisible();
    if (bVisState != bVisible)
    {
      it->second->SetChartVisibility(bVisible);
      if (bVisible == false) 
      {
        if (it->second->IsChartSelected())
        {
          m_bSelEnabled = false;
          m_chartSelIdx = -1;
          it->second->SelectChart(false);
        }

        if (m_dataViewChartIdx == chartIdx)
        {
          if (m_pChartDataView != NULL)
          {
            if (IsWindow(m_pChartDataView->m_hWnd))
              m_pChartDataView->DestroyWindow();
            delete m_pChartDataView;
            m_pChartDataView = NULL;
          }
          UpdateDataViewPnts(-1, 0, PointD(0.0, 0.0), false);
        }
      }
      else if (m_chartSelIdx == -1)
        m_bSelEnabled = false;
    
      if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
        UpdateContainerWnds();
    
      return true;
    }
  }
  
  return false;
}

int CChartContainer::ToggleChartVisibility(int chartIdx)
{
  MAP_CHARTS::iterator it = m_mapCharts.find(chartIdx);
  if (it == m_mapCharts.end())
    return -1;
  CChart* chartPtr = it->second;
  bool bVisible = chartPtr->ToggleChartVisibility();
  if (!bVisible)
  {
    if (chartPtr->IsChartSelected())
    {
      m_bSelEnabled = false;
      m_chartSelIdx = -1;
      chartPtr->SelectChart(false);
    }
    if (m_dataViewChartIdx == chartIdx)
    {
      if (m_pChartDataView != NULL)
      {
        if (IsWindow(m_pChartDataView->m_hWnd))
          m_pChartDataView->DestroyWindow();
        delete m_pChartDataView;
        m_pChartDataView = NULL;
      }
      UpdateDataViewPnts(-1, 0, PointD(0.0, 0.0), false);
    }
  }
  else if (m_chartSelIdx == -1)
    m_bSelEnabled = false;

  if (IsWindow(m_hWnd) && IsWindowVisible())
    UpdateContainerWnds();

  return bVisible ? 1 : 0;
}

std::pair<float, bool> CChartContainer::GetChartPenWidth(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return make_pair(it->second->GetPenWidth(), true);
  return make_pair(0.0f, false);
}

bool CChartContainer::UpdateChartPenWidth(int chartIdx, float penWidth, bool bRedraw)
{
  bool bRes = false;
  MAP_CHARTS::iterator it = (chartIdx == -1) ? m_mapCharts.begin() :
                                             m_mapCharts.find(chartIdx);
  for (; it != m_mapCharts.end(); ++it)
  {
    CChart* chartPtr = it->second;
    if ((chartPtr != NULL) &&((chartIdx != -1) || chartPtr->IsChartVisible()))
    { 
      if (chartPtr->GetPenWidth() != penWidth)
      {
        chartPtr->SetPenWidth(penWidth);
        bRes = true;
      }
    }
    if (chartIdx != -1)
      break;
  }

  if (bRes && bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    UpdateContainerWnds();

  return bRes;
}

std::pair<float, bool> CChartContainer::GetChartTension(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return make_pair(it->second->GetTension(), true);
  return make_pair(0.0f, false);
}

bool CChartContainer::SetChartTension(int chartIdx, float fTension, bool bRedraw)
{
  bool bRes = false;
  if (in_range(0.0f, 1.0f, fTension))
  {
    MAP_CHARTS::iterator it = (chartIdx != -1) ? m_mapCharts.find(chartIdx): m_mapCharts.begin();
    MAP_CHARTS::iterator itE = m_mapCharts.end();
    for (; it != itE; ++it)
    {
      if ((chartIdx != -1) || (it->second->IsChartVisible()))
      {
        if (it->second->GetTension() != fTension)
        {
          it->second->SetTension(fTension);
          bRes = true;
        }
      }
      if (chartIdx != -1)
        break;
    }

    if (bRes && bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
      RefreshWnd();
  }
  return bRes;
}

std::pair<bool, bool> CChartContainer::AreChartPntsAllowed(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return make_pair(it->second->ArePntsAllowed(), true);
  return make_pair(false, false);
}

bool CChartContainer::AllowChartPnts(int chartIdx, bool bAllowed, bool bRedraw)
{
  bool bRes = false;
  MAP_CHARTS::iterator it = (chartIdx != -1) ? m_mapCharts.find(chartIdx) :
                                                            m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    if ((chartIdx != -1) || (it->second->IsChartVisible()))
    {
      if (it->second->ArePntsAllowed() != bAllowed)
      {
        it->second->AllowPnts(bAllowed);
        bRes = true;
      }
    }
    if (chartIdx != -1)
      break;
  }

  if (bRes && bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    RefreshWnd();
 
  return bRes;
}

std::pair<Color, bool> CChartContainer::GetChartColor(int chartIdx) const
{
  if (chartIdx != -1)
  {
    MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
    if (it != m_mapCharts.cend())
      return make_pair(it->second->GetChartColor(), true);
  }

  return make_pair(Color::Black, false);
}

bool CChartContainer::SetChartColor(int chartIdx, Color col, bool bRedraw)
{
  bool bRes = false;

  MAP_CHARTS::iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.end())
  {
    if (it->second->GetChartColor().GetValue() != col.GetValue())
    {
      it->second->SetChartColor(col);
      if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
        UpdateContainerWnds();
      bRes = true;
    }
  }
  return bRes;
}

std::pair<DashStyle, bool> CChartContainer::GetChartDashStyle(int chartIdx) const
{
  if (chartIdx != -1)
  {
    MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
    if (it != m_mapCharts.cend())
      return make_pair(it->second->GetChartDashStyle(), true);
  }

  return make_pair(DashStyleSolid, false);
}

bool CChartContainer::SetChartDashStyle(int chartIdx, DashStyle dashStyle, bool bRedraw)
{
  bool bRes = false;
  MAP_CHARTS::iterator it = (chartIdx != -1) ? m_mapCharts.find(chartIdx): m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();

  for (; it != itE; ++it)
  {
    if ((chartIdx != -1) || it->second->IsChartVisible())
    {
      if (it->second->GetChartDashStyle() != dashStyle)
      {
        it->second->SetDashStyle(dashStyle);
        bRes = true;
      }
    }
    if (chartIdx != -1)
      break;
  }
  if (bRes && bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    UpdateContainerWnds();

  return bRes;
}
  
int CChartContainer::HasSelectedChart(void) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.begin();
  MAP_CHARTS::const_iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    if (it->second->IsChartSelected())
      return it->first;
  }
  return -1;
}

int CChartContainer::SelectChart(int chartIdx, bool bRedraw)
{
  int chartSelIdx = -1 ;
  MAP_CHARTS::iterator it, itE = m_mapCharts.end();;
  if (m_chartSelIdx != -1)     // Deselect selected chart
  {
    it = m_mapCharts.find(m_chartSelIdx);
    if (it != itE)
      it->second->SelectChart(false);
  }

  if (m_bSelEnabled &&(chartIdx != -1) && (chartIdx != m_chartSelIdx)) // Now select the chart
  {
    it = m_mapCharts.find(chartIdx);
    if (it != itE)
    {
      it->second->SelectChart(true);
      chartSelIdx = chartIdx;          
    }
  }

  if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    RefreshWnd();      

  return chartSelIdx;
}

int CChartContainer::SelectFirstVisibleChart(bool bRedraw)
{
  int chartSelIdx = -1;
  if (!m_mapCharts.empty())
  {                             // Find first visible chart
    MAP_CHARTS::iterator it = m_mapCharts.begin(), itSel, itE = m_mapCharts.end();
    for (; it != itE; ++it)
    {
      if (it->second->IsChartVisible())
      {
        chartSelIdx = it->second->GetChartIdx(); // Visible chart is found
        if (chartSelIdx != m_chartSelIdx)        // Different first visible chart
        {
          if (m_chartSelIdx != -1)           // Deselect selected chart
          {
            itSel = m_mapCharts.find(m_chartSelIdx);
            if (itSel != itE)
              itSel->second->SelectChart(false);
          }
          it->second->SelectChart(true);
          if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
            RefreshWnd();
        }
        break;
      }
    }
  }
  return chartSelIdx;
}

int CChartContainer::SelectChartByMouseClick(CPoint pnt, bool bRedraw)
{
  if (!m_mapCharts.empty())
  {
    PointF epsF(CHART_DTPNTSZ, CHART_PNTSTRSH);
    PointF pntF(float(pnt.x), float(pnt.y));

    PointD leftPntD = m_pMatrixD->TransformToPntD(1.0, pntF - epsF);
    PointD rightPntD = m_pMatrixD->TransformToPntD(1.0, pntF + epsF);

    CChart* chartPtr;
    MAP_CHARTS::iterator it = m_mapCharts.begin();
    MAP_CHARTS::iterator itE = m_mapCharts.end();

    V_CHARTDATAF vDataPntsF;
    
    for (; it != itE; ++it)
    {
      chartPtr = it->second;
      chartPtr->SelectChart(false);     
// Now look for sel pnts: Convert to the scteen pntsF
      V_CHARTDATAD vDataPntsD = chartPtr->m_vDataPnts;
      chartPtr->ConvertChartData(vDataPntsD, vDataPntsF, m_pMatrixD, leftPntD.X, rightPntD.X, false);
      if (vDataPntsF.size() > 0)
      {
        V_CHARTDATAF::iterator itData = find_if(vDataPntsF.begin(), vDataPntsF.end(),
                                                     in_vicinity<double>(pntF, epsF));
        if (itData != vDataPntsF.end())
        {
          return SelectChart(chartPtr->GetChartIdx(), bRedraw);
        }
      } // If point is found
    } // For
  } // Map not empty
  return -1;
}

// Chart Y axis precision
int CChartContainer::GetChartPrecisionY(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return it->second->GetPrecisionY();
  return INT_MAX;
}

bool CChartContainer::SetChartPrecisionY(int chartIdx, int precisionY, bool bRedraw)
{
  bool bRes = false;
  MAP_CHARTS::iterator it = (chartIdx != -1) ? m_mapCharts.find(chartIdx) :
                                                               m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    if ((chartIdx != -1) || it->second->IsChartVisible())
    {
      if (it->second->GetPrecisionY() != precisionY)
      {
        it->second->SetPrecisionY(precisionY);
        bRes = true;
      }
    }
    if (chartIdx != -1)
      break;
  }

  if (bRes && bRedraw)
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);    
    if (chartIdx == -1)
      chartIdx = m_dataViewChartIdx;
    UpdateDataView(chartIdx, F_NODATACHANGE);
  }
  return bRes;
}

string_t CChartContainer::GetChartName(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return it->second->GetChartName();
  return string_t(_T(""));     // Not a name at all
}

bool CChartContainer::ChangeChartName(int chartIdx, string_t name, bool bRedraw)
{
  MAP_CHARTS::iterator itCh = m_mapCharts.find(chartIdx);
  MAP_CHARTS::iterator it = m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();

  if (itCh != itE)                    // Chart is found
  {
    name = NormalizeString(name, STR_MAXLEN, STR_NORMSIGN); 

    for (; it != itE; ++it)           // Check for uniquiness
    {
      if (it->second->GetChartName() == name)
        return false;
    }
    itCh->second->SetChartName(name); // Assign the new nane
    
    if (bRedraw)
      UpdateContainerWnds(chartIdx);

    return true;
  }
  return false;
}

// Chart Y data name
string_t CChartContainer::GetChartYValName(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return it->second->GetAxisYName();
  return string_t(_T(""));    
}

bool CChartContainer::ChangeChartYValName(int chartIdx, string_t nameY, bool bRedraw)
{
  bool bRes = false;

  nameY = NormalizeString(nameY, STR_MAXLEN, STR_NORMSIGN);
  MAP_CHARTS::iterator it = chartIdx != -1 ? m_mapCharts.find(chartIdx) :
                                                                      m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    if ((chartIdx != -1) || it->second->IsChartVisible())
    {
      string_t oldNameY = it->second->GetAxisYName();
      if (oldNameY != nameY)
      {
        it->second->SetAxisYName(nameY);
        bRes = true;
      }
    }
    if (chartIdx != -1)
      break;
  }
   
  if (bRes &&bRedraw)
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);
    UpdateDataView(chartIdx != -1 ? chartIdx : m_dataViewChartIdx, F_NODATACHANGE);
  }
  return bRes;
}

// Function to get Y val strung function
val_label_str_fn CChartContainer::GetLabYValStrFnPtr(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return it->second->GetLabYValStrFnPtr();
  return NULL;
}

bool CChartContainer::SetLabYValStrFn(int chartIdx, val_label_str_fn pLabYValStrFn, bool bRedraw)
{
  bool bRes = false;

  MAP_CHARTS::iterator it = chartIdx != -1 ? m_mapCharts.find(chartIdx): m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();

  for (; it != itE; ++it)
  {
    if ((chartIdx != -1) || it->second->IsChartVisible())
    {
      val_label_str_fn oldFnPtr = it->second->GetLabYValStrFnPtr();
      if (oldFnPtr != pLabYValStrFn)
      {
        it->second->SetLabYValStrFn(pLabYValStrFn);
        bRes = true;
      }
    }
    if (chartIdx != -1)
      break;
  }
   
  if (bRes && bRedraw)
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);
    UpdateDataView(chartIdx != -1 ? chartIdx : m_dataViewChartIdx, F_NODATACHANGE);
  }  
 
  return bRes;
}

CChart* CChartContainer::GetChart(int chartIdx) const
{
  CChart* chartPtr = NULL;

  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    chartPtr = it->second;
  return chartPtr;
}

V_CHARTDATAD* CChartContainer::GetChartData(int chartIdx)
{
  CChart* chartPtr = GetChart(chartIdx);
  if (chartPtr != NULL)
    return &(chartPtr->m_vDataPnts);
  return NULL;
}

CChart* CChartContainer::FindChartByName(string_t nameStr) const
{
  CChart* chartPtr = NULL;
  MAP_CHARTS::const_iterator it = m_mapCharts.cbegin();
  MAP_CHARTS::const_iterator itE = m_mapCharts.cend();
  for (; it != itE; ++it)
  {
    if ( it->second->GetChartName() == nameStr)
    {
      chartPtr = it->second;
      break;
    }
  }
  return chartPtr;
}

int CChartContainer::GetChartIdx(string_t chartName) const
{
  CChart* chartPtr = FindChartByName(chartName);
  if (chartPtr != NULL)
    return chartPtr->GetChartIdx();
  return -1;
}

std::pair<double, bool>CChartContainer::GetChartLocScaleY(int chartIdx) const
{
  MAP_CHARTS::const_iterator it = m_mapCharts.find(chartIdx);
  if (it != m_mapCharts.cend())
    return make_pair(it->second->GetLocScaleY(), true);
  return make_pair(0.0, false);
}

bool CChartContainer::UpdateChartLocScaleY(int chartIdx, double locScY, bool bRedraw)
{
  bool bRes = false;
  MAP_CHARTS::iterator it = (chartIdx != -1) ? m_mapCharts.find(chartIdx) :
                                                                m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    CChart* chartPtr = it->second;
    double locScaleY = chartPtr->GetLocScaleY();
    if (locScaleY != locScY)
    {
      chartPtr->SetLocScaleY(max(0.1, locScY));
      bRes = true;
    }
    if (chartIdx != -1)
      break;
  }

  if (bRes && bRedraw)
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);
    else
      RefreshWnd();
  }
  return bRes;
}

bool CChartContainer::ZoomMoveContainer(double startX, double endX)
{
  double fStartX = GetInitialStartX();
  double fEndX   = GetInitialEndX();

// Don't zoom/move an empty space: one of the points should be inside
  if (in_range(fStartX, fEndX, startX) || in_range(fStartX, fEndX, endX))
  {
    m_vHist.push_back(PAIR_POS(m_startX, m_endX));
    m_startX = min(startX, endX);
    m_endX   = max(startX, endX);
    m_chMode = MODE_ZOOM;   // All prepared but not rendered
    return true;
  }
  return false;
}

CH_MODE CChartContainer::ZoomContainer(double startX, double endX, bool bRedraw)
{
  if (true == ZoomMoveContainer(startX, endX))
  {
    if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    {
      if (m_bTracking && IsLabWndExist(false))
        UpdateDataLegend(true);
      else
        RefreshWnd();
      m_chMode = MODE_ZOOMED; // Old extensionX is in m_dqHist
    }
    else
      m_chMode = MODE_ZOOM;   // Wait for redrawing
  }
  else
    m_chMode = m_vHist.empty() ? MODE_FULLX : MODE_ZOOMED;

  return m_chMode;
 }

CH_MODE CChartContainer::MoveContainerCharts(double deltaX, bool bRedraw)
{
  if (true == ZoomMoveContainer(m_startX - deltaX, m_endX - deltaX))
  {
    if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    {
      if (m_bTracking && IsLabWndExist(false))
        UpdateDataLegend(true);
      else
        RefreshWnd();
      m_chMode = MODE_ZOOMED; // Old extensionX is in m_dqHist
    }
    else
      m_chMode = MODE_ZOOM;   // Wait for redrawing
  }
  else
    m_chMode = m_vHist.empty() ? MODE_FULLX : MODE_ZOOMED;

  return m_chMode;
 }

LRESULT CChartContainer::SendNotification(UINT code, int chartIdx)
{
  NMCHART nmchart;
  nmchart.hdr.hwndFrom = m_hWnd;
  nmchart.hdr.idFrom = GetDlgCtrlID();
  nmchart.hdr.code = code;
  nmchart.chartIdx = chartIdx;

  switch (code)
  {
  case CODE_VISIBILITY: nmchart.bState = IsChartVisible(chartIdx);            break;
  case CODE_SHOWPNTS:   nmchart.bState = AreChartPntsAllowed(chartIdx).first; break;
  case CODE_EXTX:       nmchart.minX   = GetStartX();
                        nmchart.maxX = GetEndX();
                        break;
  default:              return 0;
  }

  CWnd* parentPtr = (CWnd*)GetParent();
  if (parentPtr != NULL)
    return parentPtr->SendMessage(WM_NOTIFY, WPARAM(nmchart.hdr.hwndFrom), LPARAM(&nmchart));
  return 0;
}

// Chart interface functions
int CChartContainer::AddChart(bool bVisible, bool bShowPnts, string_t label,
                               string_t labelY, int precisionY,
                               DashStyle dashStyle, float penWidth, float tension, 
                               Color colChart, V_CHARTDATAD& vData, bool bRedraw)
{
  if (vData.size() < 3)
    return -1;

  int chartIdx = GetMaxChartIdx() + 1;
  bool bAddIdx = false;
  if (!label.empty())
  {
    label = NormalizeString(label, STR_MAXLEN, STR_NORMSIGN);
    CChart* twinPtr = FindChartByName(label);
    if (twinPtr != NULL)
      bAddIdx = true;
  }
  else
  {
    label = string_t(_T("Cnart"));
    bAddIdx = true;
  }

  if (bAddIdx)
  {
    _TCHAR buffer_t[64];
    _itot_s(chartIdx, buffer_t, 10);  // Chart idx to string
    string_t idxStr(buffer_t);
    label += string_t(_T("_")) + string_t(buffer_t);
  }

  CChart* chartPtr = new CChart;   

  chartPtr->SetChartAttr(bVisible, bShowPnts, chartIdx, label, labelY, 
                                    precisionY, dashStyle, penWidth, tension, colChart);

// Now transfer data and set min max values
  chartPtr->m_vDataPnts.assign(vData.begin(), vData.end());
// It is cheaper to sort right away than to look for max/min x and sort later if needed
  sort(chartPtr->m_vDataPnts.begin(), chartPtr->m_vDataPnts.end(), less_pnt<double, false>());
  chartPtr->m_vDataPnts.shrink_to_fit();

  double minValX = chartPtr->m_vDataPnts.front().X;
  double maxValX = chartPtr->m_vDataPnts.back().X;

  // Find min and max Y
  PAIR_ITS pair_minmaxY = 
       minmax_element(chartPtr->m_vDataPnts.begin(), chartPtr->m_vDataPnts.end(), 
                                                            less_pnt<double, true>());
  double minValY = pair_minmaxY.first->Y;
  double maxValY = pair_minmaxY.second->Y;

// Save in the CChart
  chartPtr->SetMinValX(minValX);
  chartPtr->SetMaxValX(maxValX);
  chartPtr->SetMinValY(minValY);
  chartPtr->SetMaxValY(maxValY);

// Just in case: idx is unique for this container
  if (m_mapCharts.insert(MAP_CHARTS::value_type(chartPtr->GetChartIdx(), chartPtr)).second == false)
  {
    delete chartPtr;
    return -1;
  }

// Now update the container's min maxes, saving the history of X
  UpdateExtX(minValX, maxValX);
  UpdateExtY(minValY, maxValY);

  if (IsWindow(m_hWnd) && m_bTracking && IsLabWndExist(false))
    PrepareDataLegend(m_dataLegPntD, m_epsX, m_pDataWnd->m_mapLabs, m_mapSelPntsD, true);

  if (bRedraw && IsWindow(m_hWnd)&&IsWindowVisible())
    UpdateContainerWnds(-1, true);

  return chartIdx;
}

// Overload for time series
int CChartContainer::AddChart(bool bVisible, bool bShowPnts, string_t label, 
                              string_t labelY, int precisionY, DashStyle dashStyle, 
                              float penWidth, float tension, 
                              Color colChart, std::vector<double>& vTmSeries, 
                              bool bRedraw)
{
  if (vTmSeries.size() < 3)
    return -1;

  V_CHARTDATAD vData(vTmSeries.size());
  transform(vTmSeries.begin(), vTmSeries.end(), vData.begin(), time_series_to_pnt<double>());

  return AddChart(bVisible, bShowPnts, label, labelY, precisionY, 
                                dashStyle, penWidth, tension, colChart, vData, bRedraw);
}

int CChartContainer::AddChart(bool bVisible, bool bShowPnts, string_t label, 
                string_t labelY, int precisionY, Gdiplus::DashStyle dashStyle, 
                float penWidth, float tension, Gdiplus::Color colChart, 
                std::vector<std::pair<double, double> >& vXYData, bool bRedraw)
{
  if (vXYData.size() < 3)
    return -1;

  V_CHARTDATAD vData(vXYData.size());
  transform(vXYData.begin(), vXYData.end(), vData.begin(), xy_to_pnt<double>());

  return AddChart(bVisible, bShowPnts, label, labelY, precisionY, 
                                dashStyle, penWidth, tension, colChart, vData, bRedraw);
}

int CChartContainer::AddChart(bool bVisible, bool bShowPnts, string_t label, string_t labelY,
                int precisionY, Gdiplus::DashStyle dashStyle, float penWidth, float tension, 
                Gdiplus::Color colChart, std::vector<double>& vX, std::vector<double>& vY, 
                bool bRedraw)
{
  size_t sizeX = vX.size();
  size_t sizeY = vY.size();
  if ((sizeX != sizeY) || (sizeX < 3))
    return -1;

  V_CHARTDATAD vData(sizeX);
  for (size_t i = 0; i < sizeX; ++i)
    vData[i] = PointD(vX[i], vY[i]);
  return AddChart(bVisible, bShowPnts, label, labelY, precisionY, 
                                dashStyle, penWidth, tension, colChart, vData, bRedraw);
}

// Appends points to the end of chart data vector. Discards all new points with X < m_endX
bool CChartContainer::AppendChartData( int chartIdx, V_CHARTDATAD& vData, bool bUpdate)
{
  MAP_CHARTS::iterator it = m_mapCharts.find(chartIdx); 
  if (it != m_mapCharts.end())
  {
    V_CHARTDATAD vLocData(vData);
// Sort; it is cheaper than search for min/max X and check whether they are at start and end
    sort(vLocData.begin(), vLocData.end(), less_pnt<double, false>());

    CChart* chartPtr = it->second;  // Append data vector
    chartPtr->AppendChartData(vLocData);
    if (IsLabWndExist(false))     // Might update m_pMatrix if m_dataLegPntD is in X range
      PrepareDataLegend(m_dataLegPntD, m_epsX, m_pDataWnd->m_mapLabs, m_mapSelPntsD, true);  

    UpdateDataView(chartIdx, F_APPEND);

    if (bUpdate)                    // Update and display if only one chart has to be processed
    {
      UpdateExtX(chartPtr->GetMinValX(), chartPtr->GetMaxValX()); // New startX/endX are at the front dqHist

      std::pair<double, double> pair_minmaxY = GetMinMaxExtY();
      m_minExtY = pair_minmaxY.first;
      m_maxExtY = pair_minmaxY.second;

// Allways show full extent Y, but keep the leg windows and the history
      if (IsWindow(m_hWnd) && IsWindowVisible())
      {
        if (IsLabWndVisible(false))
          UpdateDataLegend(true);
        else
          RefreshWnd();
      }
    }
    return true;
  }
  return false;
}

// Overload for time series
bool CChartContainer::AppendChartData(int chartIdx,  std::vector<double>& vTmSeries,  bool bUpdate)
{
  if (vTmSeries.size() < 3)
    return false;

  V_CHARTDATAD vData(vTmSeries.size());
  transform(vTmSeries.begin(), vTmSeries.end(), vData.begin(), time_series_to_pnt<double>());

  return AppendChartData(chartIdx, vData, bUpdate);
}

// Overload for vector of pairs
bool CChartContainer::AppendChartData(int chartIdx,  
              std::vector<std::pair<double, double> >& vXYData, bool bUpdate)
{
  if (vXYData.size() < 3)
    return false;

  V_CHARTDATAD vData(vXYData.size());
  transform(vXYData.begin(), vXYData.end(), vData.begin(), xy_to_pnt<double>());

  return AppendChartData(chartIdx, vData, bUpdate);
}

// Overload for vectors of doubles
bool CChartContainer::AppendChartData(int chartIdx,  
              std::vector<double>& vX, std::vector<double>& vY, bool bUpdate)
{
  size_t sizeX = vX.size();
  size_t sizeY = vY.size();
  if ((sizeX != sizeY) || (sizeX < 3))
    return false;

  V_CHARTDATAD vData(sizeX);
  for (size_t i = 0; i < sizeX; ++i)
    vData[i] = PointD(vX[i], vY[i]);

  return AppendChartData(chartIdx, vData, bUpdate);
}

bool CChartContainer::ReplaceChartData(int chartIdx, V_CHARTDATAD& vData, bool bClip, 
                                                            bool bVerbose, bool bRedraw)
{
  if (vData.size() < 3)    // We don't recognize two pnts as a curve
    return false;

  MAP_CHARTS::iterator mapIt = m_mapCharts.find(chartIdx);
  if (mapIt == m_mapCharts.end())
    return false;                           // Wrong Chart ID

 // Warning and reset if (bReset)                               
  if (bVerbose && (AfxMessageBox(CString("You will lost all CChart data. Reset anyway?"), 
                                                      MB_YESNO|MB_ICONQUESTION) == IDNO)) 
    return false;

// Sort input vector; it is cheaper than check first whether it was sorted
  V_CHARTDATAD vLocData(vData);
  sort(vLocData.begin(), vLocData.end(), less_pnt<double, false>());

  CChart* chartPtr = mapIt->second;
  
// Define what take from vData
  V_CHARTDATAD::iterator itStart = vLocData.begin();
  V_CHARTDATAD::iterator itEnd = vLocData.end();

  if (bClip) // Asked to keep the new range inside the old startX, endX
  {
    PAIR_ITS pair_its = find_border_pnts(vLocData.begin(), vLocData.end(),
       not_inside_range<double, false>(GetInitialStartX(), GetInitialEndX()));
    itStart = pair_its.first;
    itEnd   = pair_its.second;
    if (itEnd != vLocData.end())
      ++itEnd;
  }

  chartPtr->m_vDataPnts.assign(itStart, itEnd);  
  chartPtr->m_vDataPnts.shrink_to_fit();

// Set new min/max
  double startX = chartPtr->m_vDataPnts.front().X;
  double endX   = chartPtr->m_vDataPnts.back().X;
  chartPtr->SetMinValX(startX);
  chartPtr->SetMaxValX(endX);

// Update scales - indirect
  std::pair< V_CHARTDATAD::iterator, V_CHARTDATAD::iterator> pair_minmaxY = 
    minmax_element(chartPtr->m_vDataPnts.begin(), chartPtr->m_vDataPnts.end(), 
                                                      less_pnt<double, true>());
  
  chartPtr->SetMinValY(pair_minmaxY.first->Y);   
  chartPtr->SetMaxValY(pair_minmaxY.second->Y);
  chartPtr->SetLocScaleY(1.0);

// Update m_startX,m_endX, min and max Y

  UpdateExtX(startX, endX); // New startX/endX are at the front dqHist
  std::pair<double, double> pair_minmaxExt = GetMinMaxExtY();
  m_minExtY = pair_minmaxExt.first;
  m_maxExtY = pair_minmaxExt.second;
// Correct the sel data points
  if (IsLabWndExist(false))     // Might update m_pMatrix if m_dataLegPntD is in X range
    PrepareDataLegend(m_dataLegPntD, m_epsX, m_pDataWnd->m_mapLabs, m_mapSelPntsD, true);  

  UpdateDataView(chartIdx, F_REPLACE);

  if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
  {
    if (IsLabWndExist(false))
      UpdateDataLegend(false);  // m_pMatrix is set already
    else
      RefreshWnd();
  }     

  return true;
}

// Overload for time series
bool CChartContainer::ReplaceChartData(int chartIdx,  std::vector<double>& vTmSeries, 
                                                              bool bVerbose, bool bRedraw)
{
  if (vTmSeries.size() < 3)
    return false;

  V_CHARTDATAD vData(vTmSeries.size());
  transform(vTmSeries.begin(), vTmSeries.end(), vData.begin(), time_series_to_pnt<double>());

  return ReplaceChartData(chartIdx, vData, bVerbose, bRedraw);
}

// Overload for vector of pairs
bool CChartContainer::ReplaceChartData(int chartIdx,  
            std::vector<std::pair<double, double> >& vXYData, bool bVerbose, bool bRedraw)
{
  if (vXYData.size() < 3)
    return false;

  V_CHARTDATAD vData(vXYData.size());
  transform(vXYData.begin(), vXYData.end(), vData.begin(), xy_to_pnt<double>());

  return ReplaceChartData(chartIdx, vData, bVerbose, bRedraw);
}
// Overload for vectors of doubles
bool CChartContainer::ReplaceChartData(int chartIdx,  
            std::vector<double>& vX, std::vector<double>& vY, bool bVerbose, bool bRedraw)
{

  size_t sizeX = vX.size();
  size_t sizeY = vY.size();
  if ((sizeX != sizeY) || (sizeX < 3))
    return false;

  V_CHARTDATAD vData(sizeX);
  for (size_t i = 0; i < sizeX; ++i)
    vData[i] = PointD(vX[i], vY[i]);

  return ReplaceChartData(chartIdx, vData, bVerbose, bRedraw);
}

void CChartContainer::ResetCharts(void)  
{
// Restore startX, endX
  m_startX = GetInitialStartX();
  m_endX   = GetInitialEndX();
  m_vHist.clear();

// Restore visibility
  MAP_CHARTS::iterator it = m_mapCharts.begin(), itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    CChart* chartPtr = it->second;
    if (!chartPtr->IsChartVisible())
      chartPtr->ToggleChartVisibility();
  }

  m_chMode = MODE_FULLX;

  UpdateContainerWnds(-1, true);
}

void CChartContainer::ResetChartContainer(bool bResetCols)
{
  CChartContainer* pContainer = new CChartContainer;

// Set defaula colors
  if (bResetCols)
  {
    m_colBkground     = pContainer->GetBkColor();
    m_colBorder       = pContainer->GetBorderColor();
    m_colAxis         = pContainer->GetAxesColor();
    m_colZoomBorder   = pContainer->GetZoomBorderColor();
    m_colLegPnts      = pContainer->GetLegPntsColor();
    m_colDataViewPnts = pContainer->GetDataViewPntsColor(); 
    m_colLegBorder    = pContainer->GetLegBkgndColor();
    m_colLegBorder    = pContainer->GetLegBorderColor();
    m_colGrid         = pContainer->GetGridColor();
  }

//  Set default values     
  m_minExtY           = pContainer->GetMinY();
  m_maxExtY           = pContainer->GetMaxY(); 
  m_startX            = pContainer->GetStartX();
  m_endX              = pContainer->GetEndX(); 
  m_axOffsXF          = pContainer->GetAxisOffsetX();
  m_axXPos            = pContainer->GetXAxisPos();
  m_axYPos            = pContainer->GetYAxisPos(); 
  m_scX               = pContainer->GetScX();
  m_scY               = pContainer->GetScY();;
  m_bTracking         = false;
  m_dataLegPntD       = pContainer->GetDataLegPntD();
  m_bSelEnabled       = false;
  m_bAxBoundaries     = false;
  m_chartSelIdx       = -1;
  m_dataViewChartIdx  = -1; 
  m_precision         = pContainer->GetContainerPrecisionX();
  m_tmpZX             = 0.0;
  m_chMode            = pContainer->GetChMode();

  delete pContainer;

// Clear maps and vectors
  if (!m_mapCharts.empty())
  {
    MAP_CHARTS::iterator it = m_mapCharts.begin();
    MAP_CHARTS::iterator itE = m_mapCharts.end();
    for (; it != itE; ++it)
      delete it->second;
    m_mapCharts.clear();
  }
  m_mapDataViewPntsD.clear();
  m_vHist.clear();
// Reset the matrix
  delete m_pMatrixD;
  m_pMatrixD = NULL;
// Get rid of the legend wnds
  m_pDataWnd = DestroyLabWnd(m_pDataWnd);  // Deletes m_mapSelPntsD, reset m_dataLegPntD
  m_pLegWnd  = DestroyLabWnd(m_pLegWnd);
// Clear the data view
  if (m_pChartDataView != NULL)
  {
    if (IsWindow(m_pChartDataView->m_hWnd))
        m_pChartDataView->DestroyWindow();
      delete m_pChartDataView;
      m_pChartDataView = NULL;
  }
}

bool CChartContainer::TruncateChart(int chartIdx, double startX, double endX, 
                                                          bool bUpdateExt, bool bRedraw)
{
  if (chartIdx == -1)
    return false;

  if (startX >= endX)     // Invalid parameters
    return false;

  if ((GetInitialStartX() >= startX)&&(GetInitialEndX() <= endX))  // Nothing to truncate
    return false;

  startX = max(startX, GetInitialStartX());
  endX   = min(endX, GetInitialEndX());

  size_t chartCnt = 0;
  MAP_CHARTS::iterator it = (chartIdx == -1) ? m_mapCharts.begin() :
                                                      m_mapCharts.find(chartIdx);
  for (; it != m_mapCharts.end(); ++it)
  {
    CChart* chartPtr = it->second;  // Append data vector
    bool bRes = chartPtr->TruncateChartData(startX, endX);
    if (bRes)
      ++chartCnt;

    if (chartIdx > -1)
      break;
  }

  if (chartCnt == 0)
    return false;

  if (bUpdateExt) // Change range X, keep range Y, keep the history
  {
    PAIR_DBLS pair_ext = GetMinMaxExtX();
    if (m_vHist.size() > 0)
      m_vHist[0] = pair_ext;
    else
    {
      m_startX = pair_ext.first;
      m_endX   = pair_ext.second;
    }

// Update extY
    pair_ext = GetMinMaxExtY();
    m_minExtY = pair_ext.first;
    m_maxExtY = pair_ext.second;
  }

  if (m_dataViewChartIdx == chartIdx)
    UpdateDataView(chartIdx, F_TRUNCATE);

  if (m_bTracking && IsLabWndExist(false))    // Will calculate and set the new translation matrix
    PrepareDataLegend(m_dataLegPntD, m_epsX, m_pDataWnd->m_mapLabs, m_mapSelPntsD, NULL);

// Now redraw
  if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
  {
    if (IsLabWndVisible(false))
      UpdateDataLegend(false);    // The matrix was already changed 
    else
      RefreshWnd();               // The matrix will changed in OnPaint()
  }
  return true;
}

bool CChartContainer::RemoveChart(int chartIdx, bool bCorrectMinMax, bool bRedraw)
{
  MAP_CHARTS::iterator mapIt = m_mapCharts.find(chartIdx);
  if (mapIt == m_mapCharts.end())
    return false;

  if (m_dataViewChartIdx == chartIdx)
  {
    if (m_pChartDataView != NULL)
      DestroyChartDataView();
  }

  CChart* chartPtr = mapIt->second;

  delete chartPtr;
  m_mapCharts.erase(chartIdx);

  if (m_mapCharts.empty())      
  {
    m_pDataWnd = DestroyLabWnd(m_pDataWnd);
    m_pLegWnd  = DestroyLabWnd(m_pLegWnd);
    ResetChartContainer();
  }
  else
  {
    if (bCorrectMinMax)
    {
      PAIR_DBLS pair_minmaxX = GetMinMaxExtX();
      PAIR_DBLS pair_minmaxY = GetMinMaxExtY();

      if (m_vHist.size() > 0)
        m_vHist[0] = pair_minmaxX;
      else
      {
        m_startX  = pair_minmaxX.first;
        m_endX    = pair_minmaxX.second;
      }
      m_minExtY = pair_minmaxY.first;
      m_maxExtY = pair_minmaxY.second;
    }

    if (IsLabWndExist(false))
      PrepareDataLegend(m_dataLegPntD, m_epsX, m_pDataWnd->m_mapLabs, m_mapSelPntsD, bCorrectMinMax);
  }
 
  if (bRedraw && IsWindow(m_hWnd) && IsWindowVisible())
    UpdateContainerWnds();

  return true;
}

// Legend functions
CDataWnd* CChartContainer::SetDataLegPntD(PointD pntD, bool bRedraw)
{
  if (in_range(m_startX, m_endX, pntD.X))
  {
    m_dataLegPntD = pntD;
    if (bRedraw)
    {
      m_bTracking = true;
      ShowDataLegend(); // Sets m_pDataWnd
    }
  }
  return m_pDataWnd;
}

// The function is called from OnLButtonDown; the container is visible
CDataWnd* CChartContainer::ShowDataLegend(CPoint pnt)
{
  if (m_mapCharts.empty())            // No charts, no legends
    return m_pDataWnd;     

  if (m_pDataWnd == NULL)             // Need to allocate memory
  {
    CDataWnd* pDataWnd = new CDataWnd(m_colLegBkgnd, m_colLegBorder);
    if (pDataWnd == NULL)
      return NULL;
    m_pDataWnd = pDataWnd;
  }

// After PrepareDataLegend mapSelData might have some points out the range Y, but mapLabs might be empty  
  if (PrepareDataLegend(pnt, m_pDataWnd->m_mapLabs, m_mapSelPntsD) > 0)
  {                                   // Create dataWnd if needed, and calculate dataWnd rects
    m_pDataWnd->CreateLegend(this, pnt, true);
    RefreshWnd();
    m_pDataWnd->ShowLegendWnd();      // Garanteed m_pDataWnd exists
  }
  else
    m_pDataWnd = DestroyLabWnd(m_pDataWnd);

  return m_pDataWnd;
}

CDataWnd* CChartContainer::ShowDataLegend(void)
{
  if (!IsWindow(m_hWnd)||!IsWindowVisible())  // Means container wnd was created and is visible
    return m_pDataWnd;
  if (m_mapCharts.empty())                    // No charts, no legends
    return m_pDataWnd;     

  CDataWnd* pDataWnd = NULL;
  if (m_pDataWnd == NULL)
  {
    pDataWnd = new CDataWnd(m_colLegBkgnd, m_colLegBorder);
    if (pDataWnd == NULL)
      return NULL;
    m_pDataWnd = pDataWnd;
  }

// After this point mapSelData might have some points out the range Y, but mapLabs might be empty  
  PrepareDataLegend(m_dataLegPntD, CHART_DTPNTSZ/m_pMatrixD->m_scX, m_pDataWnd->m_mapLabs, 
                                                                                  m_mapSelPntsD, false);
// Create dataWnd if needed, and calculate dataWnd rects
  PointF dataLegPntF = m_pMatrixD->TransformToPntF(1.0, m_dataLegPntD); 
  CPoint pnt = CPointFromPntF(dataLegPntF);
  m_pDataWnd->CreateLegend(this, pnt, true);
  RefreshWnd();
  m_pDataWnd->ShowLegendWnd();  // Garanteed m_pDataWnd exists

  return m_pDataWnd;
}

CDataWnd* CChartContainer::ShowNamesLegend(void)
{
  if (!IsWindow(m_hWnd)||!IsWindowVisible()) // Means m_pMatrix is genetated
    return m_pLegWnd;
  if (m_mapCharts.empty())            // No charts, no legends
    return m_pLegWnd;     

  CDataWnd* pLegWnd = NULL;
  if (m_pLegWnd == NULL)
  {
    pLegWnd = new CDataWnd(m_colLegBkgnd, m_colLegBorder);
    if (pLegWnd == NULL)
      return NULL;
    m_pLegWnd = pLegWnd;
  }

// After this point mapLabs might not be empty: other way all charts are not visible  
  PrepareNamesLegend(m_pLegWnd->m_mapLabs);
// Create dataWnd if needed, and calculate dataWnd rects
  m_pLegWnd->CreateLegend(this, CPoint(0, 0), false);
  RefreshWnd();
  m_pLegWnd->ShowLegendWnd();  // Garanteed m_pLegWnd exists

  return m_pLegWnd;
}

void CChartContainer::UpdateDataLegend(bool bChangeMatrix)
{
  if (m_bTracking && IsLabWndExist(false))
  {
    if (bChangeMatrix)
      SetTransformMatrixD();
    UpdateDataLegend(m_mapSelPntsD, m_pDataWnd->m_mapLabs);  

    RefreshWnd(); // UpdateDataLegend draws dataWnd; RefreshWnd redraws the region excluding dataWnd
  }
}

void CChartContainer::CopyDataLegendWnd(void)
{
  if (m_pMatrixD == NULL)
    SetTransformMatrixD();

  if (m_pDataWnd == NULL)
    m_pDataWnd = new CDataWnd(m_colLegBkgnd, m_colLegBorder);
  bool bRes = false;
  if ((m_pDataWnd != NULL)&&!IsWindow(m_pDataWnd->m_hWnd))  
    bRes = m_pDataWnd->CreateLegend(this);
  if (bRes)
    UpdateDataLegend(false);
}

// True for info on m_pLegWnd, false on m_pDataWnd
bool CChartContainer::IsLabWndVisible(bool bLeg) const
{
  CDataWnd* pDataWnd = bLeg ? m_pLegWnd : m_pDataWnd;
  if (pDataWnd != NULL)
   return (IsWindow(pDataWnd->m_hWnd) && pDataWnd->IsWindowVisible())? true : false;

  return false;
}

bool CChartContainer::IsLabWndExist(bool bLeg) const
{
  CDataWnd* pDataWnd = bLeg ? m_pLegWnd : m_pDataWnd;
  if (pDataWnd != NULL)
   return IsWindow(pDataWnd->m_hWnd)? true : false;

  return false;
}

CDataWnd* CChartContainer::DestroyLabWnd(CDataWnd* pWnd)
{
  if (pWnd != NULL)
  {
    if (IsWindow(pWnd->m_hWnd))   
    {
      pWnd->DestroyWindow();
      if (pWnd->IsDataLeg())
      {
        m_mapSelPntsD.clear();
        m_dataLegPntD = PointD(DBL_MAX, DBL_MAX);
// To remove selected data line
        RefreshWnd();
      }
    } // IsWindow
    delete pWnd;      // Here mapLegs goes out
    pWnd = NULL;
  }
  return pWnd;
}

void CChartContainer::UpdateContainerWnds(int chartIdx, bool bMatrix, DATAVIEW_FLAGS dataChange)
{
  if (m_bTracking && IsLabWndExist(false))
  {
    UpdateDataLegend(bMatrix);
  }
  else
    RefreshWnd();

  if (IsLabWndExist(true))
    ShowNamesLegend();

  UpdateDataView(chartIdx, dataChange);
}

void CChartContainer::HideLabWnds(bool bRedraw)
{
  if (IsLabWndVisible(false))
  {
    m_pDataWnd->ShowWindow(SW_HIDE);
    if (bRedraw)    // To remove sel line
      RefreshWnd();
  }
 if (IsLabWndVisible(true))
    m_pLegWnd->ShowWindow(SW_HIDE);
}

void CChartContainer::DestroyChartDataView(void)
{
  if (m_pChartDataView != NULL)
  {
    if (IsWindow(m_pChartDataView->m_hWnd))
      m_pChartDataView->DestroyWindow();
    delete m_pChartDataView;
    m_pChartDataView = NULL;
  }
  m_mapDataViewPntsD.clear();
}

// Protected helpers

// To get max chart Idx in the mapCharts
int CChartContainer::GetMaxChartIdx(void) const
{
  int chartIdx = -1;
  if (!m_mapCharts.empty())
  {
    MAP_CHARTS::const_reverse_iterator rIt = m_mapCharts.rbegin();
    chartIdx = rIt->first;
  }

  return chartIdx;
}

// To define invalidated region
CRect CChartContainer::GetExtXClientRect(double startX, double endX)
{
  CRect clRect;
  GetClientRect(&clRect);

  PointD startPntD(startX, 0.0);
  PointD endPntD(endX, 0.0);
  TRANSFORM_TO_PNTF trPntD(1.0, m_pMatrixD);

  PointF startPntF = trPntD(startPntD);
  PointF endPntF = trPntD(endPntD);

  CPoint startPnt = CPointFromPntF(startPntF);
  CPoint endPnt   = CPointFromPntF(endPntF);

  return CRect(startPnt.x, clRect.top, endPnt.x, clRect.bottom);
}

void CChartContainer::MousePntToPntD(CPoint pnt, PointD& dataPntD, MatrixD* pMatrixD)
{
  ENSURE(pMatrixD != NULL);

  PointF pntF(static_cast<float>(pnt.x), static_cast<float>(pnt.y));
// Return to a native coordinates; in most uses pntD.Y is not important, hence 1.0
  dataPntD = pMatrixD->TransformToPntD(1.0, pntF);

  // Test for NAN
  if (dataPntD.X != dataPntD.X)
    dataPntD.X = DBL_MAX;
  if (dataPntD.Y != dataPntD.Y)
    dataPntD.Y = DBL_MAX;
}

// Calculate scaleX, scaleY
double CChartContainer::UpdateScaleX(const float drawRectWidth, double startX, double endX)
{
  double scX = 1.0;
  if (!m_mapCharts.empty())
  {
    double rangeX = endX - startX;
    scX = drawRectWidth/rangeX;
  }
  return scX;
}

double CChartContainer::UpdateScaleY(const float drawRectHeight, double minY, double maxY)
{
  double scY = 1.0;
  if (!m_mapCharts.empty())
  {
    double rangeY = maxY - minY;
    scY = drawRectHeight/rangeY;
  }
  return scY;
}

// Calculates scX, scY but does not set m_scX, m_scY, leaves it to the user
PAIR_DBLS CChartContainer::UpdateScales(const RectF drawRectF, 
                                  double startX, double endX, double minY, double maxY)
{
  if (m_mapCharts.empty())
    return make_pair(1.0, 1.0);

  RectF dRF = drawRectF;
  dRF.Inflate(0.0f, -0.1f*drawRectF.Height); // Reserve 20%
  double scX = UpdateScaleX(dRF.Width, startX, endX);
  double scY = UpdateScaleY(dRF.Height, minY, maxY);
  return make_pair(scX, scY);
}

// Zooming on mouse left click + MK_SHIFT: first click
double CChartContainer::BeginZooming(CPoint pnt, bool bDraw)
{
  m_chMode = MODE_ZOOMING;
 
  PointD mousePntD;
//  MousePntToPntD(pnt, mousePntD, m_pMatrix);
  MousePntToPntD(pnt, mousePntD, m_pMatrixD);
  double startX = mousePntD.X;
  if (bDraw)
    DrawZoomBorder(pnt);
  return startX;
}

// Zooming on mouse left click + MK_SHIFT: the second click
double CChartContainer::PrepareZoom(CPoint pnt, double startX, bool bDraw)
{
  PointD mousePntD;
  MousePntToPntD(pnt, mousePntD, m_pMatrixD);
  double endX = mousePntD.X;

  if (true == ZoomMoveContainer(startX, endX)) // old m_endX, m_startX in m_dqHist
  {
    if (bDraw)
      DrawZoomBorder(pnt);
  }
  else
    RefreshWnd();   // To erase the first zoom border

  return endX;
}

// Restores the mode to ZOOM or FULLX after interruption between the first and second zoom clicks
CH_MODE CChartContainer::RestoreMode(void)
{
  if (m_chMode == MODE_ZOOMING)
  {
    if (m_vHist.empty())
      m_chMode = MODE_FULLX;
    else
      m_chMode = MODE_ZOOMED;
  }
  return m_chMode;
}

// Protected data and names legend functions    
size_t CChartContainer::PrepareDataLegend(CPoint pnt, MAP_LABSTR& mapLabels, 
                                                    MAP_SELPNTSD& mapSelPntsD)
{
  PointD dataLegPntD;
  MousePntToPntD(pnt, dataLegPntD, m_pMatrixD);

  double epsX = CHART_DTPNTSZ/m_pMatrixD->m_scX; // m11 is the scX

  size_t mapLabelsSize = PrepareDataLegend(dataLegPntD, epsX, mapLabels, mapSelPntsD, false);
  if (mapSelPntsD.size() > 0)
  {
    m_dataLegPntD = dataLegPntD;
    m_epsX        = epsX;
  }
  else 
    m_dataLegPntD = PointD(DBL_MAX, DBL_MAX);

  return mapLabelsSize;
}

size_t CChartContainer::PrepareDataLegend(PointD origPntD, double epsX, MAP_LABSTR& mapLabels,
                                                     MAP_SELPNTSD& mapSelPntsD, bool bChangeMatrix)
{
  if (bChangeMatrix)
    SetTransformMatrixD();

  mapSelPntsD.clear();
  mapLabels.clear();

 if (in_range(m_startX, m_endX, origPntD.X)) // Make sense if pnt was sent by a programmer
  {
    CRect clientR;              // Get Y-boundaries in the data space
    GetClientRect(&clientR);
    CPoint topPnt(0, clientR.top);
    CPoint bottomPnt(0, clientR.bottom);
    PointD topPntD;
    PointD bottomPntD;
    MousePntToPntD(topPnt, topPntD, m_pMatrixD);
    MousePntToPntD(bottomPnt, bottomPntD, m_pMatrixD);

//Search the chart map
    PointD selPntD;
    
    MAP_CHARTS::iterator it = m_mapCharts.begin();
    MAP_CHARTS::iterator itE = m_mapCharts.end();
    
    for (; it != itE; ++it)
    {
      CChart* chartPtr= it->second;
      bool bVisible = chartPtr->IsChartVisible();

      PAIR_ITNEAREST pair_res = 
                            chartPtr->GetNearestPointD(origPntD, epsX, selPntD);
      int nSel = pair_res.second;
      if (nSel > 0)    // Might be out of the client rect
      {
        V_CHARTDATAD::iterator itSel = pair_res.first;
        for (int i = nSel; i > 0; --i)
        {
          PointD selPntD = *itSel;
          double selY = selPntD.Y * chartPtr->GetLocScaleY();
          int chartIdx = chartPtr->GetChartIdx();
          mapSelPntsD.insert(MAP_SELPNTSD::value_type(chartIdx, selPntD));
          if (bVisible && in_range(bottomPntD.Y, topPntD.Y, selY))    // Remember, Y was reversed
          {
            TUPLE_LABEL tuple_res = chartPtr->GetSelValString(selPntD, m_labelX, m_precision, m_pLabValStrFnPtr);
            mapLabels.insert(MAP_LABSTR::value_type(chartIdx , tuple_res));
          }
          if (i > 1)
          --itSel;
        }
      }
    }
    
    return mapLabels.size();
  }
  return 0;
}

size_t CChartContainer::PrepareNamesLegend(MAP_LABSTR& mapLabels)
{
  if (m_mapCharts.empty())
    return 0;

  if (!mapLabels.empty())
    mapLabels.clear();

  TUPLE_LABEL tuple_res;
  MAP_CHARTS::iterator it = m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)  
  {
    bool bSelected = it->second->GetVisibleChartNameAndVisuals(tuple_res);
    if (bSelected)
      mapLabels.insert(MAP_LABSTR::value_type(it->first, tuple_res));
  }
  return mapLabels.size();
}

// Refreshes label string map according to charts' visibility and X range
size_t CChartContainer::UpdateDataLegend(MAP_SELPNTSD& mapSelPntsD, MAP_LABSTR& mapLabStr)
{
  mapLabStr.clear();
  if (!mapSelPntsD.empty()&& in_range(m_startX, m_endX, m_dataLegPntD.X))
  {
    CRect clRect;
    GetClientRect(&clRect);
    CPoint pntLimYL(0, clRect.bottom);
    CPoint pntLimYR(0, clRect.top);
    PointD pntLimYLD, pntLimYRD;
    MousePntToPntD(pntLimYL, pntLimYLD, m_pMatrixD);
    MousePntToPntD(pntLimYR, pntLimYRD, m_pMatrixD);

    MAP_SELPNTSD::iterator itSel = mapSelPntsD.begin();
    MAP_SELPNTSD::iterator itSelE = mapSelPntsD.end();
    while(itSel != itSelE)
    {
      int chartIdx = itSel->first;
      CChart* chartPtr = GetChart(chartIdx);
      if (chartPtr != NULL)
      {
        if (chartPtr->IsChartVisible())
        {
          PointD selPntD = itSel->second;
          if (in_range(m_startX, m_endX, selPntD.X)&&
            in_range(pntLimYLD.Y, pntLimYRD.Y, selPntD.Y*chartPtr->GetLocScaleY()))
          {
            TUPLE_LABEL tuple_res = chartPtr->GetSelValString(selPntD, m_labelX, m_precision, m_pLabValStrFnPtr);
            mapLabStr.insert(MAP_LABSTR::value_type(chartIdx, tuple_res));
          }
        }
        ++itSel;
      }
      else
       itSel = mapSelPntsD.erase(itSel);
    }
  }

  CPoint origPnt(-1, -1); // Not used on empty mapLabStr
  if (!mapLabStr.empty())
  {
    PointF origPntF = m_pMatrixD->TransformToPntF(1.0, m_dataLegPntD);
    origPnt = CPointFromPntF(origPntF);
  }
// Recalc dataWnd window rects and show data wnd or will hide it
  m_pDataWnd->UpdateDataLegend(mapLabStr, this, origPnt);
  return mapLabStr.size();
}

// Gets axis X pos AX_YPOS (first) and axis X float offsY(second)
PAIR_XAXPOS CChartContainer::GetXAxisPos(RectF rChartF, double minY, double maxY)
{
  AX_YPOS axYPos;
  float vertOffs;

// Let's begin with X axis
  if ((minY < 0)&&(maxY > 0))
  {
    double offsXY = rChartF.Height*fabs(maxY)/(maxY - minY);
    vertOffs = rChartF.GetTop() + float(offsXY);
    axYPos = MIDDLE;        // Y pos of Axis X, somewhere inside the y range
  }
  else if (minY >= 0)
  {
    vertOffs = rChartF.GetBottom();
    axYPos = BOTTOM;        // Y pos of Axis X at bottom of the draw rect
  }
  else if (maxY <= 0)
  {
    vertOffs = rChartF.GetTop();
    axYPos = TOP;           // At the top of the draw rect
  }

  return make_pair(axYPos, vertOffs);
}

// Gets axis Y pos AX_XPOS (first) and axis Y float offsX(second)
PAIR_YAXPOS CChartContainer::GetYAxisPos(RectF rChartF, double startX, double endX)
{
// Y axis
  AX_XPOS axXPos;;
  float horzOffs;

  if ((startX < 0)&&(endX > 0))
  {
    double offsYX = rChartF.Width*fabs(startX)/(endX - startX);
    horzOffs = rChartF.GetLeft() + float(offsYX);
    axXPos = BETWEEN;      // Somewhere between minX, maxX
  }
  else if (startX >= 0)
  {
    horzOffs = rChartF.GetLeft();
    axXPos = LEFT;         // At the left of the draw rect
  }
  else if (endX <= 0)
  {
    horzOffs = rChartF.GetRight();
    axXPos = RIGHT;       // At the right side of the draw rect
  }

  return make_pair(axXPos, horzOffs);
}

bool CChartContainer::ShowDataView(CChart* chartPtr, bool bClearMap, bool bRefresh)
{
  if (m_pChartDataView == NULL)
    m_pChartDataView = new CChartDataView;

  if (m_pChartDataView != NULL)
  {
    if (!IsWindow(m_pChartDataView->m_hWnd))
    {
      CRect parentWndRect;
      GetParent()->GetWindowRect(&parentWndRect); // App main dlg window

      CRect workRect;
      SystemParametersInfo(SPI_GETWORKAREA, NULL, &workRect, 0);

      int leftX  = parentWndRect.right + DV_SPACE;
      int rightX = leftX + DV_RECTW;
      int topY   = parentWndRect.top - DV_SPACE;
      int bottomY = topY + DV_RECTH;

      CRect dataViewRect(leftX, topY, rightX, bottomY);
      CRect interRect;
      interRect.IntersectRect(&dataViewRect, workRect);
      if (interRect != dataViewRect)
      {
        dataViewRect.right = workRect.right - DV_SPACE;
        dataViewRect.left = max(dataViewRect.right - DV_RECTW, workRect.left + DV_SPACE);
        dataViewRect.top = workRect.top + DV_SPACE;
        dataViewRect.bottom = min(dataViewRect.top + DV_RECTH, workRect.bottom - DV_SPACE);
      }

      BOOL bRes =  m_pChartDataView->CreateEx(0,
                                 AfxRegisterWndClass(CS_HREDRAW|CS_VREDRAW|CS_SAVEBITS),
                                 _T("Chart Data View"), 
                                 WS_POPUPWINDOW|WS_CAPTION|WS_MINIMIZEBOX|WS_VISIBLE,
                                 dataViewRect.left, dataViewRect.top, dataViewRect.Width(), dataViewRect.Height(),
                                 NULL, 
                                 NULL,
                                 NULL);
      if (!bRes)
      {
        delete m_pChartDataView;
        m_pChartDataView = NULL;
        return false;
      }
    }
    else if (m_pChartDataView->IsIconic())
      m_pChartDataView->ShowWindow(SW_RESTORE);

    int chartIdx = chartPtr->GetChartIdx();
    m_pChartDataView->ShowWaitMessage(chartIdx, chartPtr->m_vDataPnts.size());
    m_pChartDataView->InitParams(chartPtr, bClearMap, this);

    if (m_dataViewChartIdx != chartIdx)
    {
      m_dataViewChartIdx = chartIdx;
      bClearMap =true;
    }

    if (bClearMap)
    {
      m_mapDataViewPntsD.clear();
      if (bRefresh)
        RefreshWnd();
    }
  }
  return true;
}


bool CChartContainer::UpdateDataView(int chartIdx, DATAVIEW_FLAGS flagsData)
{
  if ((m_pChartDataView != NULL) && (IsWindow(m_pChartDataView->m_hWnd)))
  {
    if ((chartIdx != -1) && (chartIdx == m_dataViewChartIdx))
    {
      CChart* chartPtr = GetChart(chartIdx);
      if (chartPtr != NULL)
        return m_pChartDataView->UpdateParams(chartPtr, flagsData);
    }
  }
  return false;
}

size_t CChartContainer::UpdateDataViewPntsDMap(void)
{
  if (m_pChartDataView != NULL)
    m_mapDataViewPntsD = m_pChartDataView->m_mapSelCells;
  else
    m_mapDataViewPntsD.clear();

  return m_mapDataViewPntsD.size();
}

// This returns chartPtr->m_loScY as a first, and deltaLocScY as a second; delta = 1% of Y range
PAIR_DBLS CChartContainer::GetLocScYDelta(CChart* chartPtr)
{
  if (IsWindow(m_hWnd))
  {
    CRect clR = GetExtXClientRect(m_startX, m_endX);
    double deltaY = 0.01*clR.Height()/m_scY;
    double locScY = chartPtr->GetLocScaleY();
    double maxY = max(fabs(chartPtr->GetMaxValY()), fabs(chartPtr->GetMinValY()));
    double deltaK = deltaY/maxY;
    return make_pair(locScY, deltaK);
  }
  return make_pair(1.0, 0.0);
}

void CChartContainer::MoveChartsByArrowKeys(UINT nChar)
{
  m_vHist.push_back(make_pair(m_startX, m_endX));
  double deltaX = 0.0;
  if (nChar == VK_RIGHT)
    deltaX = MOVE_DELTAX*(m_endX - m_startX);
  else if (nChar == VK_LEFT)
    deltaX = -MOVE_DELTAX*(m_endX - m_startX);
  else 
    return;

  MoveContainerCharts(deltaX, true);
}

void CChartContainer::ChangeLocScaleYByArrowKeys(UINT nChar)
{
  if (m_bSelEnabled &&(m_chartSelIdx > -1))
  {
    MAP_CHARTS::iterator it = m_mapCharts.find(m_chartSelIdx);
    if (it != m_mapCharts.end())
    {
      CChart* chartPtr= it->second;
      if (chartPtr->IsChartSelected() && chartPtr->IsChartVisible())
      {
        std::pair<double, double> pair_scY = GetLocScYDelta(chartPtr);
        double locScY = pair_scY.first;
        double deltaK = pair_scY.second;

        switch (nChar)
        {
        case VK_DOWN: locScY = max(locScY - deltaK, DELTA_LOCY);      break; 
        case VK_UP:   locScY += deltaK;                               break; 
        case VK_PRIOR: locScY += 10.0*deltaK;                         break; 
        case VK_NEXT:locScY = max(locScY - 10.0*deltaK, DELTA_LOCY);  break;
        }

        chartPtr->SetLocScaleY(locScY);  
        
        if (IsLabWndExist(false))
        {
          if (m_pDataWnd->IsWindowVisible())
            m_pDataWnd->ShowWindow(SW_HIDE);
          SetTimer(1, 500, NULL);             // To fight flicker
        }
        RefreshWnd();
     }
   }
 }
}

// Draw functions
PointF CChartContainer::DrawAxes(RectF rClientF,double startX, double endX, double minY, 
                              double maxY, Color colAxis, Graphics* grPtr, float dpiRatio)
{
  Pen axisPen(colAxis, AXIS_PENWIDTH*dpiRatio);  

  PointF pntStartF, pntEndF, pntCoordZeroF;
// Let's begin with X axis

  PAIR_XAXPOS pair_axXYPos = GetXAxisPos(rClientF, minY, maxY);
  m_axYPos = pair_axXYPos.first;
  pntStartF.Y = pntEndF.Y = pair_axXYPos.second;
  pntStartF.X = rClientF.GetLeft() - DR_SPACE*dpiRatio;   // Don't worry: draw area is clipped
  pntEndF.X   = rClientF.GetRight() + DR_SPACE*dpiRatio;
  grPtr->DrawLine(&axisPen, pntStartF, pntEndF);

  pntCoordZeroF.Y = pntStartF.Y;

// Y axis
  PAIR_YAXPOS pair_axYXPos = GetYAxisPos(rClientF, startX, endX);
  m_axXPos = pair_axYXPos.first;
  pntStartF.X = pntEndF.X = pair_axYXPos.second;
  pntStartF.Y = rClientF.GetBottom() + DR_SPACE*dpiRatio;
  pntEndF.Y   = rClientF.GetTop() - DR_SPACE*dpiRatio;
  grPtr->DrawLine(&axisPen, pntStartF, pntEndF);

  pntCoordZeroF.X = pntStartF.X;

  m_axOffsXF = pntCoordZeroF.Y; 
  return pntCoordZeroF;
}

void CChartContainer::DrawBkground(RectF rClientF, Color colBkground, Color colBorder, 
                                                                  Graphics* grPtr)
{
  SolidBrush bkBrush(colBkground);
  grPtr->FillRectangle(&bkBrush, rClientF);

  Pen borderPen(colBorder);
  grPtr->DrawRectangle(&borderPen, rClientF);

  if (this == GetFocus())
  {
    rClientF.Inflate(-3.0f, -3.0f);
    borderPen.SetDashStyle(DashStyleDash);
    grPtr->DrawRectangle(&borderPen, rClientF);
  }
}

bool CChartContainer::DrawSelPntsD(CChart* chartPtr, const MAP_SELPNTSD& mapSelPntsD, 
                GraphicsPath* grPathPtr, Graphics* grPtr, float dpiRatio)
{
  if (!in_range(m_startX, m_endX, m_dataLegPntD.X))
    return false;
  if (mapSelPntsD.empty() || (chartPtr == NULL))
    return false;

// Get screen pntsF for the given chart
  V_CHARTDATAD vDataPntsD(mapSelPntsD.size());
  V_CHARTDATAF vDataPntsF;

  int chartIdx = chartPtr->GetChartIdx();
// Find first entry
  MAP_SELPNTSD::const_iterator itSel = mapSelPntsD.find(chartIdx); 
  if (itSel != m_mapSelPntsD.cend())                              // The chart has selected points                         
  {
    size_t count = m_mapSelPntsD.count(chartIdx);
    for (size_t i = 0; i < count; ++i)
    {
      PointD selPntD = itSel->second;
      if (in_range(m_startX, m_endX, selPntD.X))              // Sometimes it will draw out of clip area
      {
        PointF selPntF = m_pMatrixD->TransformToPntF(chartPtr->GetLocScaleY(), selPntD);
        vDataPntsF.push_back(selPntF);
      }
      ++itSel;
    }
// Draw
    if (!vDataPntsF.empty())
    {
      V_CHARTDATAF::iterator it = vDataPntsF.begin(), itE = vDataPntsF.end();
  
      for (; it != itE; ++it)
      {
        PointF selPntF = *it;
        RectF rPntF = RectFFromCenterF(selPntF, dpiRatio*CHART_DTPNTSZ, dpiRatio*CHART_DTPNTSZ);
        rPntF.Inflate(dpiRatio*1.5f, dpiRatio*1.5f);
        grPathPtr->AddEllipse(rPntF);
      }

      Pen selPen(m_colLegPnts, dpiRatio*1.0f);
      grPtr->DrawPath(&selPen, grPathPtr);

      grPathPtr->Reset();
      return true;
    }   // For vDataPntsF
  }
  return false;
}

bool CChartContainer::DrawDataViewPntsD(const MAP_SELPNTSD& mapDataViewPntsD, double startX, 
                          double endX, GraphicsPath* grPathPtr, Graphics* grPtr, float dpiRatio)
{
  if (mapDataViewPntsD.empty())
    return false;
  if (NULL == GetVisibleChart(m_dataViewChartIdx))
    return false;
  
  CChart* chartPtr = GetChart(m_dataViewChartIdx);
  if ((chartPtr == NULL)||!chartPtr->IsChartVisible())
    return false;

// Extract dataPoints from the map into a vector
  V_CHARTDATAD vDataPntsD(mapDataViewPntsD.size());
  transform(mapDataViewPntsD.begin(), mapDataViewPntsD.end(), vDataPntsD.begin(), 
                                                       get_map_value<int, PointD>());
// Exclude elements out of the range and convert to screen pntsF
  V_CHARTDATAF vDataPntsF;
  if (!chartPtr->ConvertChartData(vDataPntsD, vDataPntsF, m_pMatrixD, startX, endX, false))
    return false;

  vDataPntsF.shrink_to_fit();

// Finally, draw
  V_CHARTDATAF::iterator it = vDataPntsF.begin();
  V_CHARTDATAF::iterator itE = vDataPntsF.end();
  for (; it != itE; ++it)
  {
    RectF rPntF = RectFFromCenterF(*it, dpiRatio*(CHART_DTPNTSZ + 2.0f), 
                                                    dpiRatio*(CHART_DTPNTSZ+ 2.0f));
    grPathPtr->AddEllipse(rPntF);
  }
  
  Pen selPen(m_colDataViewPnts, dpiRatio*2.0f);
  grPtr->DrawPath(&selPen, grPathPtr);
  grPathPtr->Reset();

  return true;
}

void CChartContainer::DrawSelLine(RectF clRF, Graphics* grPtr, float dpiRatio)
{
  if (m_mapSelPntsD.empty())
    return;
  if (NULL == GetVisibleChart(-1))
    return;
  if (IsLabWndExist(false)||(dpiRatio != 1.0f)) // On printout there is no dataWnd
  {

    if (in_range(m_startX, m_endX, m_dataLegPntD.X))
    {  
// Get client coord of the data line
      PointF pntF = m_pMatrixD->TransformToPntF(1.0, m_dataLegPntD);
// Again, no problem with printing out of a clipping rectangle
      Pen selPen(m_colLegPnts, dpiRatio*1.0f);
      grPtr->DrawLine(&selPen, pntF.X, clRF.Y, pntF.X, clRF.GetBottom());
    }
  }
}

void CChartContainer::DrawAxXBoundaryStr(RectF boundRF, Graphics* grPtr, float dpiRatio)
{
  string_t strStartX, strEndX;

  bool bAddEqSign = m_labelX.empty() ? false : true;

  string_t strSpace(_T(" "));
  strStartX = m_labelX + strSpace + m_pLabValStrFnPtr(m_startX, m_precision, bAddEqSign);
  strEndX   = m_labelX + strSpace + m_pLabValStrFnPtr(m_endX, m_precision, bAddEqSign);

  FontFamily fontFamily(_T("Verdana"));
  Gdiplus::Font axFont(&fontFamily, 7.0f, FontStyleBold);
  grPtr->SetTextRenderingHint(TextRenderingHintAntiAliasGridFit);

// Measure strings
  boundRF.Inflate(-dpiRatio*2.0f, -dpiRatio*2.0f);
  RectF startRF, endRF;
  PointF startPntF, endPntF;

  grPtr->MeasureString(strStartX.c_str(), -1, &axFont, boundRF, &startRF); 
  grPtr->MeasureString(strEndX.c_str(), -1, &axFont, boundRF, &endRF);

  if ((m_axYPos == MIDDLE)||(m_axYPos == TOP))
  {
    startPntF.Y = m_axOffsXF + dpiRatio*DR_SPACE;
    endPntF.Y    = m_axOffsXF + dpiRatio*DR_SPACE;
  }
  else if (m_axYPos == BOTTOM)
  {
    startPntF.Y = m_axOffsXF - startRF.Height - dpiRatio*(3.0f + DR_SPACE);
    endPntF.Y   = m_axOffsXF - endRF.Height - dpiRatio*(3.0f + DR_SPACE);
  }

  if (m_axXPos == BETWEEN)
  {
    startPntF.X = startRF.X  + dpiRatio*DR_SPACE;
    endPntF.X   = boundRF.GetRight() - endRF.Width - dpiRatio*DR_SPACE;
  }
  else if (m_axXPos == LEFT)
  {
    startPntF.X = startRF.X  + dpiRatio*(3.0f + DR_SPACE);
    endPntF.X   = boundRF.GetRight() - endRF.Width - dpiRatio*DR_SPACE;
  }
  else if (m_axXPos == RIGHT)
  {
    startPntF.X = startRF.X  + dpiRatio*DR_SPACE;
    endPntF.X   = boundRF.GetRight() - endRF.Width - dpiRatio*(3.0f + DR_SPACE);
  }
  else
    return;

  startRF.X = startPntF.X;
  startRF.Y = startPntF.Y;
  endRF.X   = endPntF.X;
  endRF.Y   = endPntF.Y;

  SolidBrush axBkBrush(m_colBkground);
  SolidBrush axLabBr(m_colAxis);
  
  grPtr->FillRectangle(&axBkBrush, startRF);
  grPtr->DrawString(strStartX.c_str(), -1, &axFont, startPntF, &axLabBr);

  grPtr->FillRectangle(&axBkBrush, endRF);
  grPtr->DrawString(strEndX.c_str(), -1, &axFont, endPntF, &axLabBr);
}

// Used to draw Zoom Border out of OnPaint function
void CChartContainer::DrawZoomBorder(CPoint pnt)
{
  CDC* pDC = GetDC();
  GraphicsPath grPath;

  CRect clR;
  GetClientRect(&clR);

  CPoint tmpPnt;
  CRect nameR;
  if (IsLabWndVisible(true))
  {
    m_pLegWnd->GetWindowRect(&nameR);
    ScreenToClient(&nameR);
    tmpPnt = CPoint(pnt.x, nameR.CenterPoint().y);
    if (!nameR.PtInRect(tmpPnt))
      nameR.SetRectEmpty();
  }

  CRect dataR;
  if (IsLabWndVisible(false))
  {
    m_pDataWnd->GetWindowRect(&dataR);
    ScreenToClient(&dataR);
    tmpPnt = CPoint(pnt.x, dataR.CenterPoint().y);
    if (!dataR.PtInRect(tmpPnt))
      dataR.SetRectEmpty();
  }

  Point startPnt(pnt.x, clR.top);
  Point endX;
  if (!nameR.IsRectEmpty())
  {
    grPath.AddLine(startPnt, Point(pnt.x, nameR.top));
    grPath.StartFigure();
    startPnt.Y = nameR.bottom;
    if (!dataR.IsRectEmpty())
    {
      grPath.AddLine(startPnt, Point(pnt.x, dataR.top));
      grPath.StartFigure();
      startPnt.Y = dataR.bottom;
    }
    grPath.AddLine(startPnt, Point(pnt.x, clR.bottom));
  }
  else if (!dataR.IsRectEmpty())
  {
    grPath.AddLine(startPnt, Point(pnt.x, dataR.top));
    grPath.StartFigure();
    startPnt.Y = dataR.bottom;
  }
  grPath.AddLine(startPnt, Point(pnt.x, clR.bottom));

  Graphics gr(pDC->m_hDC);
  Pen brdPen(m_colZoomBorder);
  gr.DrawPath(&brdPen, &grPath);
//  gr.DrawLine(&brdPen, pnt.x, clR.top, pnt.x, dataR.top);//clR.bottom);
  ReleaseDC(pDC);
}

void  CChartContainer::DrawGrid(RectF boundRF, PointF coordOrigF, Graphics* grPtr, float dpiRatio)
{
  float fDeltaX = boundRF.Width/10.0f;
  float fDeltaY = boundRF.Height/10.0f;

  float limTop    = boundRF.GetTop();
  float limBottom = boundRF.GetBottom();
  float limLeft   = boundRF.GetLeft();
  float limRight  = boundRF.GetRight();

  Pen gridPen(m_colGrid, 1.0f*dpiRatio);
  gridPen.SetDashStyle(DashStyleDash);

  float xF, yF;

  yF = coordOrigF.Y + fDeltaY;
  while (yF < limBottom)
  {
    grPtr->DrawLine(&gridPen, limLeft, yF, limRight, yF);
    yF += fDeltaY;
  }

  yF = coordOrigF.Y - fDeltaY;
  while (yF > limTop)
  {
    grPtr->DrawLine(&gridPen, limLeft, yF, limRight, yF);
    yF -= fDeltaY;
  }

  xF = coordOrigF.X + fDeltaX;
  while (xF < limRight)
  {
    grPtr->DrawLine(&gridPen, xF, limTop, xF, limBottom);
    xF += fDeltaX;
  }
 
  xF = coordOrigF.X - fDeltaX;
  while (xF > limLeft)
  {
    grPtr->DrawLine(&gridPen, xF, limTop, xF, limBottom);
    xF -= fDeltaX;
  }
}

// Printing
int CChartContainer::GetScreenDpi(void)
{
  CPaintDC containerDC(this);
  int scrDpiX = containerDC.GetDeviceCaps(LOGPIXELSX);
  int scrDpiY = containerDC.GetDeviceCaps(LOGPIXELSY);
  ENSURE(scrDpiX == scrDpiY);
  return scrDpiX;
}

TUPLE_PRNLEGLAYOUT CChartContainer::CalcNamesPntLayout(RectF namesRF, 
                         const MAP_PRNDATA& mapPrnData, Graphics* grPtr, float dpiRatio)
{
// Measure strings
  float maxNamesLength = 0, maxScYLength = 0, maxValStrLength = 0;
  float maxNameXStrLength = 0, maxStrXLength = 0; 
  float maxNameYStrLength = 0, maxStrYLength = 0;

  RectF tmpRF; 

// Get max size of the printing cells
  FontFamily fontFamily(_T("Verdana"));
  Gdiplus::Font labelFont(&fontFamily, 8.0f, FontStyleBold);
  grPtr->SetTextRenderingHint(TextRenderingHintAntiAliasGridFit);

  // Measure strings
  MAP_PRNDATA::const_iterator it = mapPrnData.begin();
  MAP_PRNDATA::const_iterator itE = mapPrnData.end();

// Get max strings and str lengths
  while (it != itE)
  {
    TUPLE_PRINT tuple_chart = it->second;
// Names
    string_t tmpStr  = get<IDX_NAME>(tuple_chart);

    if (!tmpStr.empty())
    {
      grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
      if (tmpRF.Width > maxNamesLength )
        maxNamesLength = tmpRF.Width;
    }
// Scales
    tmpStr = get<IDX_SCY>(tuple_chart);
    if (!tmpStr.empty())
    {
      grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
      if (tmpRF.Width > maxScYLength)
        maxScYLength = tmpRF.Width;
    }
// Sel pnts if any
    tmpStr = get<IDX_VAL>(tuple_chart);
    if (!tmpStr.empty())  // If empty, no selected points
    {
      grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
      if (tmpRF.Width > maxValStrLength)
        maxValStrLength = tmpRF.Width;
    }

// X name
    tmpStr = get<IDX_NAMEX>(tuple_chart);
    if (!tmpStr.empty())
    {
      grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
      if (tmpRF.Width > maxNameXStrLength)
        maxNameXStrLength = tmpRF.Width;
    }

// X values
    tmpStr = get<IDX_X>(tuple_chart);
    grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
    if (tmpRF.Width > maxStrXLength)
      maxStrXLength = tmpRF.Width;

// Y name str
    tmpStr = get<IDX_NAMEY>(tuple_chart);
    if (!tmpStr.empty())
    {
      grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
      if (tmpRF.Width > maxNameYStrLength)
        maxNameYStrLength = tmpRF.Width;
    }

// Y values
    tmpStr = get<IDX_Y>(tuple_chart);
    grPtr->MeasureString(tmpStr.c_str(), -1, &labelFont, namesRF, &tmpRF);
    if (tmpRF.Width > maxStrYLength)
      maxStrYLength = tmpRF.Width;

    ++it;
  }

  // Column offsets, begin from the end
  float endX = namesRF.GetRight();
  float startX = namesRF.GetLeft();

  float valYOffsX = endX - maxStrYLength;
  float nameYStrOffsX = maxNameYStrLength == 0 ? valYOffsX : 
                                valYOffsX  - maxNameYStrLength - dpiRatio*LB_BORDERSPACE;   
  float valXOffsX = nameYStrOffsX - maxStrXLength - dpiRatio*LB_BORDERSPACE;
  float nameXStrOffsX = maxNameXStrLength == 0 ? valXOffsX :
                                valXOffsX - maxNameXStrLength - - dpiRatio*LB_BORDERSPACE;
  float valStrOffsX = maxValStrLength == 0 ? nameXStrOffsX :
                              nameXStrOffsX - maxValStrLength - - dpiRatio*LB_BORDERSPACE;
  float scYOffsX = maxScYLength == 0 ? valStrOffsX :
                                valStrOffsX - maxScYLength - dpiRatio*LB_BORDERSPACE;
  float nameStrOffsX = maxNamesLength == 0 ? scYOffsX :
                                  scYOffsX - maxNamesLength - dpiRatio*LB_BORDERSPACE;
  float bulletOffsX = maxNamesLength == 0 ? nameStrOffsX :
                          nameStrOffsX - dpiRatio*LB_BULLETSZ - dpiRatio*LB_BORDERSPACE;
// Adjust offsets:
  int stepsY = 0;
  if (bulletOffsX >= startX)
  {
    float offsShiftX = (bulletOffsX - startX)/2.0f;      // Actually startX + (bulletOffsX - startX)/2 
    bulletOffsX   -= offsShiftX;
    nameStrOffsX  -= offsShiftX;
    scYOffsX      -= offsShiftX;
    valStrOffsX   -= offsShiftX;
    nameXStrOffsX -= offsShiftX;
    valXOffsX     -= offsShiftX;
    nameYStrOffsX -= offsShiftX;
    valYOffsX     -= offsShiftX;
  }
  else
  {
    float offsShiftX = startX - bulletOffsX;
    bulletOffsX  += offsShiftX;
    nameStrOffsX += offsShiftX;
    scYOffsX     += offsShiftX;

    stepsY = STEPY_VALSTR;

    if (valStrOffsX < startX)                             // Three lines
    {
      float valWidth = max(maxStrXLength, maxStrYLength);  // Beautification
      float nmWidth  = max(maxNameXStrLength, maxNameYStrLength);
      valYOffsX      = endX - valWidth;
      nameYStrOffsX  = maxNameYStrLength == 0 ? valYOffsX :
                          valYOffsX - nmWidth - dpiRatio*LB_BORDERSPACE;
      valXOffsX      = valYOffsX;
      nameXStrOffsX  = maxNameXStrLength == 0 ? valXOffsX :
                      valXOffsX - nmWidth - dpiRatio*LB_BORDERSPACE;
      valStrOffsX    = maxValStrLength == 0 ?  nameXStrOffsX :
           min(nameXStrOffsX, nameYStrOffsX) - maxValStrLength - dpiRatio*LB_BORDERSPACE;
      stepsY |= STEPY_VALY;
    }
  }

  TUPLE_PRNLEGLAYOUT tuple_res;

  get<IDX_BULLETSTART>(tuple_res) = bulletOffsX, 
  get<IDX_NAME>(tuple_res)        = nameStrOffsX;
  get<IDX_SCY>(tuple_res)         = scYOffsX;
  get<IDX_VAL>(tuple_res)         = valStrOffsX;;
  get<IDX_NAMEX>(tuple_res)       = nameXStrOffsX;
  get<IDX_X>(tuple_res)           = valXOffsX;
  get<IDX_NAMEY>(tuple_res)       = nameYStrOffsX;
  get<IDX_Y>(tuple_res)           = valYOffsX;

  get<IDX_STEPSY>(tuple_res)      = stepsY;

  return tuple_res;
}

void CChartContainer::PrintCrossLine(PointF pntStartF, PointF pntEndF, SizeF sizeCrossF, 
                                               Pen* penPtr, Graphics* grPtr, float dpiRatio)
{
  PointF pntXF(dpiRatio*sizeCrossF.Width/2, 0.0f);
  PointF pntYF(0.0f, dpiRatio*sizeCrossF.Height/2);
  float space = dpiRatio*(sizeCrossF.Width + 1.0f);
// Lets calculate number of Crosses
  int CrossNmb = static_cast<int>(floor((pntEndF.X - pntStartF.X + dpiRatio*1.0f)/space));

  GraphicsPath grPath;
  pntStartF.X += pntXF.X;   // Move to the center of the first fugure

  for (int i = 0; i < CrossNmb; ++i)
  {
    grPath.AddLine(pntStartF - pntXF, pntStartF + pntXF);
    grPath.StartFigure();
    grPath.AddLine(pntStartF - pntYF, pntStartF + pntYF);
    grPath.StartFigure();

    pntStartF.X += space;
  }
  grPtr->DrawPath(penPtr, &grPath);  
}

size_t CChartContainer::PrintChartNames(RectF namesRF, const MAP_PRNDATA& mapPrnData, 
           const TUPLE_PRNLEGLAYOUT& tuple_prnLayout, Graphics* grPtr, float dpiRatio, size_t mapOffs)
{
// Set the font
  FontFamily fontFamily(_T("Verdana"));
  Gdiplus::Font labelFont(&fontFamily, 8.0f, FontStyleBold);
  grPtr->SetTextRenderingHint(TextRenderingHintAntiAliasGridFit);
  float fontHeight = labelFont.GetHeight(grPtr);
  float lineHeight = 1.5f*fontHeight; // In doc. units, 1/300 of inch

// Get offsets
  float lineY    = namesRF.Y;

// Get string layout; do not care whether there are val, nMEx, and nameY  strings
  float bulletStartOffsX = get<IDX_BULLETSTART>(tuple_prnLayout);
  float bulletEndX       = bulletStartOffsX + dpiRatio*LB_BULLETSZ;
  float nmStrOffsX       = get<IDX_NAME>(tuple_prnLayout);;
  float scYStrOffsX      = get<IDX_SCY>(tuple_prnLayout);
  float valStrOffsX      = get<IDX_VAL>(tuple_prnLayout);     
  float nmXOffsX         = get<IDX_NAMEX>(tuple_prnLayout);
  float valXOffsX        = get<IDX_X>(tuple_prnLayout);
  float nmYOffsX         = get<IDX_NAMEY>(tuple_prnLayout);
  float valYOffsX        = get<IDX_Y>(tuple_prnLayout);

  int stepsY             = get<IDX_STEPSY>(tuple_prnLayout);

  Pen bulletPen((ARGB)Color::Black);      // Pen Width is to set later
  SolidBrush txtBr((ARGB)Color::Black);

  MAP_PRNDATA::const_iterator it = mapPrnData.begin();
  MAP_PRNDATA::const_iterator itE = mapPrnData.end();

  advance(it, mapOffs);
  bool bName = mapOffs == 0 ? false : true;   // To begin the new page with full line
  for (; it != itE; ++it)
  {
    TUPLE_PRINT tuple_chart = it->second;     // Get strings prepared
// Set pen and brush attributes
    Color strColor = get<IDX_COL>(tuple_chart);
    bulletPen.SetColor(strColor);
    DashStyle dashStyle = get<IDX_DSTYLE>(tuple_chart);
    bulletPen.SetDashStyle(dashStyle != DashStyleCustom ? dashStyle : DashStyleSolid);
    bulletPen.SetWidth(dashStyle != DashStyleCustom ? 
                                      get<IDX_PWIDTH>(tuple_chart)*dpiRatio : dpiRatio);
    txtBr.SetColor(strColor);

    PointF pntStartF, pntEndF;

    string_t tmpStr = get<IDX_NAME>(tuple_chart); 
    if (!tmpStr.empty())            
    {
      bName = false;
      pntStartF = PointF(bulletStartOffsX, lineY + fontHeight/2.0f);
      pntEndF   = PointF(bulletEndX, lineY + fontHeight/2.0f);

      if (dashStyle != DashStyleCustom)
      {
        grPtr->DrawLine(&bulletPen, pntStartF, pntEndF);
      }
      else
      {
        PrintCrossLine(pntStartF, pntEndF, SizeF(CHART_DTPNTSZ, CHART_DTPNTSZ),  
                                                      &bulletPen, grPtr, dpiRatio);
      }
// Draw name string
      grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(nmStrOffsX, lineY), &txtBr);
// Draw scY string 
      tmpStr = get<IDX_SCY>(tuple_chart);
      grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(scYStrOffsX, lineY), &txtBr);
    }
    else if (bName)
    {
      MAP_PRNDATA::const_reverse_iterator rIt = MAP_PRNDATA::const_reverse_iterator(it);
      for (; rIt != mapPrnData.crend(); ++rIt)
      {
        TUPLE_PRINT tuple_nm = rIt->second;
        tmpStr = get<IDX_NAME>(tuple_nm);
        if (tmpStr.empty())
          continue;
        if (dashStyle != DashStyleCustom)
        {
          grPtr->DrawLine(&bulletPen, pntStartF, pntEndF);
        }
        else
        {
           PrintCrossLine(pntStartF, pntEndF, SizeF(CHART_DTPNTSZ, CHART_DTPNTSZ),  
                                                      &bulletPen, grPtr, dpiRatio);
        }
        grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(nmStrOffsX, lineY), &txtBr);
        tmpStr = get<IDX_SCY>(tuple_nm);
        grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(scYStrOffsX, lineY), &txtBr);
        bName = false;
        break;
      }
    }

    if (stepsY & STEPY_VALSTR)
      lineY += lineHeight;
    
    tmpStr = get<IDX_VAL>(tuple_chart);
    if (!tmpStr.empty())
      grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(valStrOffsX, lineY), &txtBr);
    tmpStr = get<IDX_NAMEX>(tuple_chart);
    if (!tmpStr.empty())
      grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(nmXOffsX, lineY), &txtBr);
    grPtr->DrawString(get<IDX_X>(tuple_chart).c_str(), -1, &labelFont, 
                                                          PointF(valXOffsX, lineY), &txtBr);
    if (stepsY & STEPY_VALY)
      lineY += lineHeight;

    tmpStr = get<IDX_NAMEY>(tuple_chart);
    if (!tmpStr.empty())
      grPtr->DrawString(tmpStr.c_str(), -1, &labelFont, PointF(nmYOffsX, lineY), &txtBr);
    grPtr->DrawString(get<IDX_Y>(tuple_chart).c_str(), -1, &labelFont, 
                                                              PointF(valYOffsX, lineY), &txtBr);

    ++mapOffs;
    lineY += lineHeight;

    if ((lineY + 3*lineHeight) > namesRF.GetBottom())
      return mapOffs;

  }
  return mapOffs;
}

string_t CChartContainer::PrintPageHeader(Gdiplus::RectF rGdiF, 
              const CChartContainer* pContainer, string_t tmStr, Gdiplus::Graphics* grPtr)
{
 // Form bounding rectangles
  RectF headerRF = rGdiF;
  headerRF.Height = 80.0f; // Approx. 7 mm 
  headerRF.Inflate(-200.0f, 0.0f);
  headerRF.Offset(0.0f, 250.0f);

// Draw header
  FontFamily fontFamily(_T("Verdana"));
  Gdiplus::Font headerFont(&fontFamily, 16.0f, FontStyleRegular, UnitPoint);
  grPtr->SetTextRenderingHint(TextRenderingHintAntiAliasGridFit);

  SolidBrush brHeader((ARGB)Color::Black);
  StringFormat strFormat;
  strFormat.SetAlignment(StringAlignmentNear);
  strFormat.SetLineAlignment(StringAlignmentCenter);

  grPtr->DrawString(pContainer->m_name.c_str(),-1, &headerFont,  headerRF, &strFormat, &brHeader);

  if (tmStr.empty())
  {
    // Print current local time
    CTime tmD = CTime::GetCurrentTime();
    CString strTmD = tmD.Format(_T("%a, %b %d, %Y, %H:%M"));
    tmStr = string_t(strTmD);
  }

// Draw time string
  strFormat.SetAlignment(StringAlignmentFar);
  grPtr->DrawString(tmStr.c_str(), -1, &headerFont,  headerRF, &strFormat, &brHeader);

  return tmStr;
}

void CChartContainer::PrintCharts(CChartContainer* pContainer, int scrDpiX, HDC printDC)
{
  if (printDC != NULL)
  {
    pContainer->m_bAxBoundaries = true;

    CDC* pDC = new CDC;
    pDC->Attach(printDC); 
    Graphics* grPtr = new Graphics(printDC);

    grPtr->SetPageUnit(UnitDocument);
    grPtr->SetSmoothingMode(SmoothingModeAntiAlias);
// Get dpiRatio
    float dpiRatioX = 300.0f/scrDpiX;

    RectF rGdiF;
    grPtr->GetVisibleClipBounds(&rGdiF);                 // The same as the clip rect

// Begin Printing
    pDC->StartDoc(pContainer->m_name.c_str());          // MFC functions
    pDC->StartPage();

// Draw header
    string_t tmStr;
    tmStr = pContainer->PrintPageHeader(rGdiF, pContainer, tmStr, grPtr);

// Format drawing area and draw a border
    RectF drawRF = rGdiF;   
    drawRF.Inflate(-200.0f, -200.0f);
    drawRF.Height = rGdiF.Width*0.75f;
    drawRF.Offset(0.0f, 200.0f);

    SolidBrush bkBr(pContainer->m_colBkground);
    grPtr->FillRectangle(&bkBr, drawRF);

    Pen pen((ARGB)Color::Black, 3*dpiRatioX);
    pen.SetAlignment(PenAlignmentInset);
    grPtr->DrawRectangle(&pen, drawRF);
// Set clipping region    
    grPtr->SetClip(drawRF);

// Draw charts
    if (pContainer->m_mapCharts.size() > 0)
    {
// Draw axes
      PointF pntOrigF = pContainer->DrawAxes(drawRF, pContainer->m_startX, pContainer->m_endX, 
                        pContainer->m_minExtY, pContainer->m_maxExtY, pContainer->m_colAxis, grPtr);
// Draw Grid
      pContainer->DrawGrid(drawRF, pntOrigF, grPtr, dpiRatioX);
// Set margins for drawing
      drawRF.Inflate(-DR_SPACE, -DR_SPACE);
// Update scales
      std::pair <double, double> pair_scales = 
        pContainer->UpdateScales(drawRF, pContainer->m_startX,pContainer->m_endX, 
                                                      pContainer->m_minExtY, pContainer->m_maxExtY);
      pContainer->m_scX = pair_scales.first;
      pContainer->m_scY = pair_scales.second;
// Set transform matrix
      MatrixD matrixD;
      matrixD.Translate(pntOrigF.X, pntOrigF.Y); // Add drawRF
      matrixD.Scale(float(pContainer->m_scX), float(-pContainer->m_scY));
      if ((pContainer->m_axXPos != BETWEEN)||(pContainer->m_axYPos != MIDDLE))
      {
        double translateX = (pContainer->m_axXPos == RIGHT) ? -pContainer->m_endX : 
                        (pContainer->m_axXPos == LEFT) ? -pContainer->m_startX : 0.0;
        double translateY = (pContainer->m_axYPos == BOTTOM) ? -pContainer->m_minExtY : 
                        (pContainer->m_axYPos == TOP) ? -pContainer->m_maxExtY : 0.0;
        matrixD.Translate(float(translateX), float(translateY)); 
      }
      if (pContainer->m_pMatrixD != NULL)
        delete pContainer->m_pMatrixD;
      pContainer->m_pMatrixD = matrixD.Clone();     // Can't copy: copy constructor is private

// Calculate page scale Y to display on the legend
      ENSURE (grPtr->GetPageUnit() == UnitDocument); // Sure it is: set it at the start of printing
      PointD pntStartD, pntEndD;
      CPoint pntStart(int(drawRF.X), int(drawRF.Y)), pntEnd(int(drawRF.GetRight()), int(drawRF.GetBottom()));
      pContainer->MousePntToPntD(pntStart, pntStartD, pContainer->m_pMatrixD);
      pContainer->MousePntToPntD(pntEnd, pntEndD, pContainer->m_pMatrixD);
      double pageScY = 300.0*(pntStartD.Y - pntEndD.Y)/drawRF.Height;  // 300 dpi/inch for the Doc. Mode
  
// Begin draw all visible curves or selected only
      int chartSelIdx = pContainer->HasSelectedChart();      
      MAP_LABSTR mapLabStr;
      size_t mapOffs = 0;
 
      GraphicsPath grPath;            // Will be used by the drawing routines
      MAP_CHARTS::iterator it;
      if (chartSelIdx == -1)
        it = pContainer->m_mapCharts.begin();
      else 
       it = pContainer->m_mapCharts.find(chartSelIdx);
      MAP_CHARTS::iterator itE = pContainer->m_mapCharts.end();

      MAP_PRNDATA mapPrnData;

      for (; it != itE; ++it)
      {
        CChart* chartPtr = it->second;
        if (chartPtr->IsChartVisible() == false)
          continue;
        if (chartPtr->m_vDataPnts.size() < 2)
          continue;
        chartPtr->DrawChartCurve(chartPtr->m_vDataPnts, pContainer->m_startX, pContainer->m_endX, 
                                                         pContainer->m_pMatrixD, &grPath, grPtr, dpiRatioX);
        chartPtr->GetSelValString(pageScY, pContainer->m_labelX, pContainer->m_precision, 
                                           pContainer->m_mapSelPntsD, mapPrnData, pContainer->m_pLabValStrFnPtr);
        if (pContainer->m_bTracking)
          pContainer->DrawSelPntsD(chartPtr, pContainer->m_mapSelPntsD, &grPath, grPtr, dpiRatioX);

        if (chartSelIdx > -1)
          break;
      }

      if (pContainer->m_bTracking)
        pContainer->DrawSelLine(drawRF, grPtr, dpiRatioX);

      if (pContainer->m_bAxBoundaries)
        pContainer->DrawAxXBoundaryStr(drawRF, grPtr);
  // Now print the legend: chart names, etc.
      grPtr->ResetClip();
      RectF namesRF(drawRF.X, drawRF.GetBottom() + 100.0f, drawRF.Width, 0.0f);
      namesRF.Height = rGdiF.GetBottom() - namesRF.Y; // temporary; will be set in PrintChartNames
      TUPLE_PRNLEGLAYOUT tuple_prnLayout = 
        pContainer->CalcNamesPntLayout(namesRF, mapPrnData, grPtr, dpiRatioX);
      mapOffs = pContainer->PrintChartNames(namesRF, mapPrnData, tuple_prnLayout, grPtr, dpiRatioX);
      if (mapOffs < mapPrnData.size())        // Print the rest of the legend on next pages
      {
// Get new namesRF
        namesRF = rGdiF;   
        namesRF.Inflate(-200.0f, -200.0f);
        namesRF.Height = rGdiF.Height - 300.0f;
        namesRF.Offset(0.0f, 300.0f);
// Print extra pages
        while (mapOffs < mapPrnData.size())
        {
          pDC->EndPage();
          pDC->StartPage();
          pContainer->PrintPageHeader(rGdiF, pContainer, tmStr, grPtr);
          mapOffs = pContainer->PrintChartNames(namesRF, mapPrnData, tuple_prnLayout, grPtr, 
                                                                dpiRatioX, mapOffs);
        }
      }
    }
// End printing
    pDC->EndPage();
    pDC->EndDoc();

    delete grPtr;
    pDC->Detach();
    delete pDC;
  }
}


BEGIN_MESSAGE_MAP(CChartContainer, CWnd)
  ON_WM_PAINT()
  ON_WM_GETDLGCODE()
  ON_WM_LBUTTONDOWN()
  ON_WM_LBUTTONUP()
  ON_WM_SETCURSOR()
  ON_WM_MBUTTONDOWN()
  ON_WM_MOUSEWHEEL()
  ON_WM_CONTEXTMENU()
  ON_COMMAND(IDM_CHART_LEGEND, OnMnShowLegend)
  ON_COMMAND(IDM_CHART_BOUNDARIES, OnMnShowAxBoundaries)
  ON_COMMAND(IDM_CHART_REFRESH, OnMnRefresh)
  ON_COMMAND(IDM_CHART_VIEWDATA, OnMnViewData)
  ON_COMMAND(IDM_CHART_SAVE, OnMnSaveCharts)
  ON_COMMAND(IDM_CHART_PRINT, OnMnPrintCharts)

  ON_COMMAND(IDM_CHART_ZOOMIN, OnMnZoomIn) 
  ON_COMMAND(IDM_CHART_MOVERIGHT, OnMnMoveRight)
  ON_COMMAND(IDM_CHART_MOVELEFT, OnMnMoveLeft)
  ON_COMMAND(IDM_CHART_TRACK, OnMnEnableTracking)
  ON_COMMAND(IDM_CHART_LOCSCALEY, OnMnEnableLocScale)
  ON_COMMAND(IDM_CHART_SHOWPNTS, OnMnShowPnts)
  ON_COMMAND(IDM_CHART_UNDOZOOMMOVE, OnMnUndoHistStep)
  ON_COMMAND_RANGE(IDM_CHART_FIRSTCHART, IDM_CHART_LASTCHART, OnMnShowChart)
#ifdef _TEST_
  ON_COMMAND(IDM_CHART_TEST, OnMnTest)
#endif
  ON_WM_KEYDOWN()
  ON_WM_KEYUP()
  ON_WM_TIMER()
  ON_WM_KILLFOCUS()
  ON_WM_SETFOCUS()
END_MESSAGE_MAP()

///////////////////////////////////////////////////////////////////////////////
// CChartContainer message handlers

void CChartContainer::OnPaint()
{
  CPaintDC dc(this);                              // Device context for painting
  Graphics gr(dc.m_hDC);                          // Graphics to paint

  Rect rGdi;
  gr.GetVisibleClipBounds(&rGdi);                 // The same as the clip rect

  Bitmap clBmp(rGdi.Width, rGdi.Height);          // Mem bitmap
  Graphics* grPtr = Graphics::FromImage(&clBmp);  // As memDC
  grPtr->SetSmoothingMode(SmoothingModeAntiAlias);

  RectF rGdiF = GdiRectToRectF(rGdi);
  rGdiF.Width -= 1.0f;
  rGdiF.Height -= 1.0f;
  DrawBkground(rGdiF, m_colBkground, m_colBorder, grPtr);


  if (m_mapCharts.size() > 0)
  {
    RestoreMode();  // Might be in the middle of settung zoon borders; annuls the mode
    RectF drawRF = rGdiF;   
    drawRF.Inflate(-DR_SPACE, -DR_SPACE);
// Draw axes
    PointF pntOrigF = DrawAxes(drawRF, m_startX, m_endX, m_minExtY, m_maxExtY, m_colAxis, grPtr);
// Draw a grid    
    DrawGrid(drawRF, pntOrigF, grPtr);

// Update scales
    std::pair<double, double> pair_scales = 
      UpdateScales(drawRF, m_startX, m_endX, m_minExtY, m_maxExtY);
    m_scX = pair_scales.first;
    m_scY = pair_scales.second;
// Set transform matrix
    MatrixD matrixD;
    matrixD.Translate(pntOrigF.X, pntOrigF.Y);
    matrixD.Scale(m_scX, -m_scY);
// Account for not even axes placement
    if ((m_axXPos != BETWEEN)||(m_axYPos != MIDDLE))
    {
      double translateX = (m_axXPos == RIGHT) ? -m_endX : (m_axXPos == LEFT) ? -m_startX : 0.0;
      double translateY = (m_axYPos == BOTTOM) ? -m_minExtY : (m_axYPos == TOP) ? -m_maxExtY : 0.0;
      matrixD.Translate(translateX, translateY);
    }
    if (m_pMatrixD != NULL)
      delete m_pMatrixD;
    m_pMatrixD = matrixD.Clone();     // Copy constructor is private ??

// Begin draw curves
    GraphicsPath grPath;            // Will be used actoss the drawing routines
    MAP_CHARTS::iterator it = m_mapCharts.begin();
    MAP_CHARTS::iterator itE = m_mapCharts.end();
    for (; it != itE; ++it)
    {
      CChart* chartPtr = it->second;
      if (chartPtr->IsChartVisible() == false)
        continue;
      if (chartPtr->m_vDataPnts.size() < 2)
        continue;
      chartPtr->DrawChartCurve(chartPtr->m_vDataPnts, m_startX, m_endX, 
                                                       m_pMatrixD, &grPath, grPtr);
      if (m_bTracking)              // Draw selection points if any
        DrawSelPntsD(chartPtr, m_mapSelPntsD, &grPath, grPtr);
    }

    if (m_bTracking)                // And the sel line too
      DrawSelLine(rGdiF, grPtr);
    if (!m_mapDataViewPntsD.empty())// Were selected in the data view
      DrawDataViewPntsD(m_mapDataViewPntsD, m_startX, m_endX, &grPath, grPtr);

    if (m_bAxBoundaries)            // Left and right visible X-coordinates
      DrawAxXBoundaryStr(rGdiF, grPtr);
 }
 // Transfer to the screen
  gr.DrawImage(&clBmp, rGdi);
  delete grPtr;
}

UINT CChartContainer::OnGetDlgCode()
{
  UINT nCode = CWnd::OnGetDlgCode();
  nCode |= DLGC_WANTARROWS;
  return nCode;
}

void CChartContainer::OnLButtonDown(UINT nFlags, CPoint point)
{
  SetFocus();
  if (nFlags & MK_SHIFT)
  {
    if ((m_chMode != MODE_ZOOMING) && (m_chMode != MODE_ZOOM))
      m_tmpZX = BeginZooming(point, true);    // Sets m_startX, m_chMode = MODE_ZOOMING
    else if (m_chMode == MODE_ZOOMING)
      PrepareZoom(point, m_tmpZX, true);     // Sets m_startX, m_endX, m_chMode = MODE_ZOOM
  }
  else if (nFlags & MK_CONTROL)
  {
    if (!m_bSelEnabled)
    {
      m_bSelEnabled = true;
      m_chartSelIdx = SelectChartByMouseClick(point, true);
      if (m_chartSelIdx == -1)
        m_chartSelIdx = SelectFirstVisibleChart(true);
    }
    else
    {
      m_chartSelIdx = SelectChartByMouseClick(point, false);
      if (m_chartSelIdx == -1)
        m_bSelEnabled = false;
      UpdateContainerWnds();
    }
  }
  else if (m_bTracking)
    ShowDataLegend(point);

  CWnd::OnLButtonDown(nFlags, point);
}

void CChartContainer::OnLButtonUp(UINT nFlags, CPoint point)
{
  if ((nFlags & MK_SHIFT)&&(m_chMode == MODE_ZOOM))
  {
    if (m_bTracking && IsLabWndExist(false))
      UpdateDataLegend(true);
    else
      RefreshWnd();

    m_chMode = MODE_ZOOMED;
  }

  CWnd::OnLButtonUp(nFlags, point);
}

void CChartContainer::OnMButtonDown(UINT nFlags, CPoint point)
{
  SetFocus();
  if (!m_mapCharts.empty() && (NULL != GetVisibleChart(-1))) 
  {
    if (nFlags & MK_CONTROL)
    {
      m_bSelEnabled = !m_bSelEnabled;
      if (m_bSelEnabled)
      {
        m_chartSelIdx =  SelectChartByMouseClick(point, true);
        if (m_chartSelIdx == -1)
          m_chartSelIdx = SelectFirstVisibleChart(true);
      }
      else                                    // Unselect sel chart
        m_chartSelIdx = SelectChart(-1, true);
    }
    else
    {
      m_bTracking = !m_bTracking;
      if (!m_bTracking)
      {
        m_pDataWnd = DestroyLabWnd(m_pDataWnd);
        m_dataLegPntD = PointD(DBL_MAX, DBL_MAX);
      }
    }
  }
  CWnd::OnMButtonDown(nFlags, point);
}

BOOL CChartContainer::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
  if (nFlags & MK_SHIFT)            // Move 
  {
    double deltaX = 0.0;
    if (zDelta < 0)
      deltaX = MOVE_DELTAX*(m_endX - m_startX);
    else if (zDelta > 0)
      deltaX = -MOVE_DELTAX*(m_endX - m_startX);
    else
      return CWnd::OnMouseWheel(nFlags, zDelta, pt);

    MoveContainerCharts(deltaX, true);
  }
  else if ((nFlags & MK_CONTROL) && m_bSelEnabled &&  (m_chartSelIdx != -1))
  {
    MAP_CHARTS::iterator it = m_mapCharts.find(m_chartSelIdx);
    if (it != m_mapCharts.end())
    {
      CChart* chartPtr = it->second;
      if (chartPtr->IsChartVisible() && chartPtr->IsChartSelected())
      {
        std::pair<double, double> pair_scY = GetLocScYDelta(chartPtr);
        double locScY = pair_scY.first;
        double deltaK = pair_scY.second;

        if (zDelta < 0)
          locScY = locScY = max(locScY - deltaK, DELTA_LOCY);
        else
          locScY += deltaK;

        if (IsLabWndExist(false))
        {
          if (m_pDataWnd->IsWindowVisible())
            m_pDataWnd->ShowWindow(SW_HIDE);
          SetTimer(1, 500, NULL);
        }

        chartPtr->SetLocScaleY(locScY);  
        RefreshWnd();
      }
    }
  }
  return CWnd::OnMouseWheel(nFlags, zDelta, pt);
}

void CChartContainer::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  if ((NULL != GetVisibleChart(-1))&&((nChar == VK_LEFT)||(nChar == VK_RIGHT)))    // Move
    MoveChartsByArrowKeys(nChar);
  else if ((nChar == VK_DOWN)||(nChar == VK_UP)||(nChar == VK_PRIOR)||(nChar == VK_NEXT))
    ChangeLocScaleYByArrowKeys(nChar);

  CWnd::OnKeyDown(nChar, nRepCnt, nFlags); 
}

BOOL CChartContainer::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message)
{
  if (nHitTest == HTCLIENT)
  {
    HCURSOR hCursor = NULL;
    if (m_bTracking)
      hCursor = LoadCursor(NULL, IDC_CROSS);
    else  
      hCursor = LoadCursor(NULL, IDC_ARROW);
    if (hCursor != NULL)
    {
      SetCursor(hCursor);
    }
    return TRUE;

  }
  return CWnd::OnSetCursor(pWnd, nHitTest, message);
}

void CChartContainer::OnTimer(UINT_PTR nIDEvent)
{
  if (nIDEvent == 1)        // Append
  {
    KillTimer(nIDEvent);

    if (m_bTracking && IsLabWndExist(false))
      UpdateDataLegend(true); 
  }
  else if (nIDEvent == 2)   // LocScY or window size was changed or size 
  {
    KillTimer(nIDEvent);
    UpdateContainerWnds(false);
  }
}

void CChartContainer::OnKillFocus(CWnd* pNewWnd)
{
  CWnd::OnKillFocus(pNewWnd);
  RefreshWnd();
}

void CChartContainer::OnSetFocus(CWnd* pOldWnd)
{
  CWnd::OnSetFocus(pOldWnd);
  RefreshWnd();
}

// Context menu functions
void CChartContainer::OnContextMenu(CWnd* /*pWnd*/, CPoint point)
{
  if (m_mapCharts.empty())
    return;

  CMenu popMenu;
  popMenu.CreatePopupMenu();   
  UINT chEnabledLeg = IsLabWndVisible(true) ? MF_CHECKED : MF_UNCHECKED;
  popMenu.AppendMenu(MF_ENABLED|chEnabledLeg, IDM_CHART_LEGEND, _T("Show Legend"));
  popMenu.AppendMenu(MF_SEPARATOR);

  UINT chShowBounds = m_bAxBoundaries ? MF_CHECKED : MF_UNCHECKED;
  popMenu.AppendMenu(MF_ENABLED|chShowBounds, IDM_CHART_BOUNDARIES, _T("Show Axis X boundaries"));
  popMenu.AppendMenu(MF_SEPARATOR);

  popMenu.AppendMenu(MF_ENABLED, IDM_CHART_VIEWDATA, _T("View Chart Data"));

  CMenu popSubMenu1;
  popSubMenu1.CreatePopupMenu();

  popSubMenu1.AppendMenu(MF_ENABLED, IDM_CHART_SAVE, _T("Save chart(s)"));
  popMenu.AppendMenu(MF_SEPARATOR);
  popSubMenu1.AppendMenuW(MF_ENABLED, IDM_CHART_PRINT, _T("Print charts"));

  popMenu.AppendMenuW(MF_STRING|MF_POPUP, UINT(popSubMenu1.m_hMenu), _T("Save/Print Charts"));
  popMenu.AppendMenu(MF_SEPARATOR);

  CMenu popSubMenu2;
  popSubMenu2.CreatePopupMenu();

  popSubMenu2.AppendMenu(MF_ENABLED, IDM_CHART_ZOOMIN, _T("Zoom In"));
  popSubMenu2.AppendMenu(MF_ENABLED, IDM_CHART_MOVERIGHT, _T("Move Right"));
  popSubMenu2.AppendMenu(MF_ENABLED, IDM_CHART_MOVELEFT, _T("Move Left"));

  popMenu.AppendMenuW(MF_STRING|MF_POPUP, UINT(popSubMenu2.m_hMenu), _T("Zoom/Move"));
  popMenu.AppendMenu(MF_SEPARATOR);

  if (!m_vHist.empty())
  {
    popMenu.AppendMenu(MF_ENABLED, IDM_CHART_UNDOZOOMMOVE, _T("Undo Last Zoom/Move"));
    popMenu.AppendMenu(MF_SEPARATOR);
    popMenu.AppendMenu(MF_ENABLED, IDM_CHART_REFRESH, _T("Refresh Charts"));
    popMenu.AppendMenu(MF_SEPARATOR);
  }

// If the mouse has no middle button replace middle button clicks
  if (GetSystemMetrics(SM_CMOUSEBUTTONS) < 3)
  {
    CMenu popSubMenu3;
    popSubMenu3.CreatePopupMenu();

    UINT chTracking = m_bTracking ? MF_CHECKED : MF_UNCHECKED;
    popSubMenu3.AppendMenu(MF_ENABLED|chTracking, IDM_CHART_TRACK, _T("Enable Tracking"));

    UINT chLocScale = m_bSelEnabled ? MF_CHECKED : MF_UNCHECKED;
    popSubMenu3.AppendMenu(MF_ENABLED|chLocScale, IDM_CHART_LOCSCALEY, _T("Enable Loc Scaling"));
    
    popMenu.AppendMenuW(MF_STRING|MF_POPUP, UINT(popSubMenu3.m_hMenu), _T("Track/Select"));
    popMenu.AppendMenu(MF_SEPARATOR);
  }

  UINT chShowCharts = MF_CHECKED;         // By default points are shown
  if (m_chartSelIdx != -1)
  {
    CChart* selChartPtr = GetChart(m_chartSelIdx);
    if (!selChartPtr->ArePntsAllowed())
      chShowCharts = MF_UNCHECKED;
  }

  popMenu.AppendMenu(MF_ENABLED|chShowCharts, IDM_CHART_SHOWPNTS, _T("Show/Hide Chart Pnts"));
  popMenu.AppendMenu(MF_SEPARATOR);

// Add CChart names (m_mapCharts is not empty)
  string_t mnPrefix(_T("Show "));
  string_t mnItemName;
  UINT chkFlag = MF_CHECKED;
  UINT cmdID = IDM_CHART_FIRSTCHART;
  MAP_CHARTS::iterator it = m_mapCharts.begin();
  MAP_CHARTS::iterator itE = m_mapCharts.end();
  for (; it != itE; ++it)
  {
    chkFlag = it->second->IsChartVisible() ? MF_CHECKED : MF_UNCHECKED;
    mnItemName = mnPrefix + it->second->GetChartName();//String();
    popMenu.AppendMenu(MF_ENABLED|chkFlag, cmdID + it->first, mnItemName.c_str());
  }
#ifdef _TEST_
  popMenu.AppendMenu(MF_SEPARATOR);
  popMenu.AppendMenu(MF_ENABLED, IDM_CHART_TEST, _T("Test"));
#endif
  popMenu.TrackPopupMenu(TPM_LEFTALIGN | TPM_LEFTBUTTON, point.x, point.y, this);
  popMenu.Detach();
}

void CChartContainer::OnMnShowLegend(void)
{
  if (IsLabWndVisible(true)) 
    m_pLegWnd = DestroyLabWnd(m_pLegWnd);
  else
    ShowNamesLegend();
}

void CChartContainer::OnMnShowAxBoundaries(void)
{
  m_bAxBoundaries = !m_bAxBoundaries;
  RefreshWnd();
}

void CChartContainer::OnMnRefresh(void)
{
  ResetCharts();
}

void CChartContainer::OnMnViewData(void)
{
  if (m_mapCharts.empty())
  {
    AfxMessageBox(_T("Chart Container Is Empty"));
    return;
  }

// Send chart to the data view dlg
  int chartSelIdx = HasSelectedChart();

  if (chartSelIdx == -1)
  {
    CChart* chartPtr = GetVisibleChart(-1);
    chartSelIdx = (NULL == chartPtr) ? -1 : chartPtr->GetChartIdx();
  }
  if (chartSelIdx == -1)
  {
    AfxMessageBox(_T("View Chart Data: Has no visible or selected chart"));
    return;
  }
 
  MAP_CHARTS::iterator it = m_mapCharts.find(chartSelIdx);
  if (it == m_mapCharts.end())
  {
    AfxMessageBox(_T("View Chart Data: Can't find selected chart"));
    return;
  }

  ShowDataView(it->second, true);
}

void CChartContainer::OnMnSaveCharts(void)
{
  if (S_FALSE == SaveChartData())
  {
    AfxMessageBox(_T("Save Failed\n Select the visible chart to save it\n")
                  _T("or remove the selection to save all visible charts\n")
                  _T("and try again"), MB_OK|MB_ICONERROR);
  }
}

void CChartContainer::OnMnPrintCharts(void)
{
  CPrintDialog printDlg(FALSE, PD_USEDEVMODECOPIES|PD_HIDEPRINTTOFILE|PD_NOSELECTION|PD_RETURNDC);
  printDlg.m_pd.Flags &= ~PD_SHOWHELP;
  printDlg.m_pd.nMinPage = printDlg.m_pd.nMaxPage = 1;
  printDlg.m_pd.nFromPage = printDlg.m_pd.nToPage = 1;

  if (printDlg.DoModal() == IDOK)
  {
    int scrDpiX = GetScreenDpi();
    CChartContainer* pContainer = CloneChartContainer(string_t(_T("")), true);
    PrintCharts(pContainer, scrDpiX, printDlg.GetPrinterDC());
    delete pContainer;
  }
}

void CChartContainer::OnMnZoomIn(void)  // ZoomIn from the context menu
{                                       
  double fStartZX, fEndZX;
  double extX = m_endX - m_startX;

  if (m_chMode == MODE_ZOOMING)         // Left zoom border was set
  {
    double distStartX = m_tmpZX - m_startX;
    double distEndX   = m_endX - m_tmpZX;
    if (distStartX < distEndX)
    {
      fStartZX = m_tmpZX;
      fEndZX   = m_endX - distStartX;
    }
    else if (distStartX > distEndX)
    {
      fStartZX = m_startX + distEndX;
      fEndZX   = m_tmpZX;
    }
    else
    {
      fStartZX = m_tmpZX - 0.9*extX;
      fEndZX   = m_tmpZX + 0.9*extX;
    }
  }
  else
  {
    fStartZX = m_startX + 0.1*extX;
    fEndZX   = m_endX   - 0.1*extX;
  }

 ZoomContainer(fStartZX, fEndZX, true);
 SendNotification(CODE_EXTX);
}

void CChartContainer::OnMnMoveRight(void)
{
// Move right
  double deltaX = MOVE_DELTAX*(m_endX - m_startX);
  MoveContainerCharts(deltaX, true);
  SendNotification(CODE_EXTX);
}

void CChartContainer::OnMnMoveLeft(void)
{
// Move left
  double deltaX = -MOVE_DELTAX*(m_endX - m_startX);
  MoveContainerCharts(deltaX, true);
  SendNotification(CODE_EXTX);
}

void CChartContainer::OnMnUndoHistStep(void)
{
  UndoHistStep(true);
  SendNotification(CODE_EXTX);
}

void CChartContainer::OnMnEnableTracking(void)
{
  m_bTracking = !m_bTracking;
  if (!m_bTracking)
  {
    m_pDataWnd = DestroyLabWnd(m_pDataWnd);
    m_dataLegPntD = PointD(DBL_MAX, DBL_MAX);
  }
}

void CChartContainer::OnMnEnableLocScale(void)
{
  if (NULL != GetVisibleChart(-1))
  {
    m_bSelEnabled = !m_bSelEnabled;
    if (m_bSelEnabled)
      m_chartSelIdx = SelectFirstVisibleChart(true);
    else 
      m_chartSelIdx = SelectChart(-1, true);
  }
  else
  {
    m_bSelEnabled = false;
    m_chartSelIdx = -1;   // Deselect selected chart if any exists
  }
}

void CChartContainer::OnMnShowPnts(void)
{
  CChart* chartPtr = NULL;
  if (m_chartSelIdx == -1)
  {
    chartPtr = GetVisibleChart(m_chartSelIdx);
    if (chartPtr != NULL)
    {
      m_bSelEnabled = true;
      m_chartSelIdx = SelectChart(chartPtr->GetChartIdx());
    }
    else
    {
      AfxMessageBox(_T("Select a chart first"), MB_ICONERROR|MB_OK);
      return;
    }
  }
  
  chartPtr = GetChart(m_chartSelIdx);
  if (chartPtr != NULL)
  {
    bool bShow = chartPtr->ArePntsAllowed();
    chartPtr->AllowPnts(!bShow);
    UpdateContainerWnds(false);
    SendNotification(CODE_SHOWPNTS, m_chartSelIdx);
  }
}

void CChartContainer::OnMnShowChart(UINT nCmdID)
{
  int chartIdx = nCmdID - IDM_CHART_FIRSTCHART;
  ToggleChartVisibility(chartIdx);
  SendNotification(CODE_VISIBILITY, chartIdx);
}

#ifdef _TEST_

void CChartContainer::OnMnTest(void)
{
// SetContainerPrecision(10);  
// SetAxisXName(string_t(_T("Distance")));
// SetLabXValStrFn(GetXLabNamedValStr, true);
//  SetChartVisibility(0, true, true);
//  UpdateChartPenWidth(-1, 4.0f, true);
//  SetChartTension(-1, 0.5f, true);
//  AllowChartPnts(-1, false, true);
  SetChartColor(0, (ARGB)Color::Magenta);
//  SetChartDashStyle(1, DashStyleDash, true);
//  SelectChart(-1, true);
//    SetChartPrecisionY(-1, 10, false);
//  ChangeChartName(0, string_t(_T("New Sin Wave")), true);
//  SetChartPrecisionY(-1, 10, false);
//  ChangeChartYValName(1, string_t(_T("WaterLevel, m")), false);
//  SetLabYValStrFn(1, GetXLabNamedValStr, false);
//  int chartIdx = 1;
//  V_CHARTDATAD* pData = GetChartData(chartIdx);
//  V_CHARTDATAD::iterator itStart = find_if(pData->begin(), pData->end(), 
//                                              greater_or_equal<double, false>(8.0));
//  V_CHARTDATAD vData;;
//  vData.assign(pData->begin(), pData->end());
//  TruncateChart(0, -10.0, 10.0, true, false);
//  TruncateChart(1, -10.0, 10.0, true, false);
//  UpdateContainerWnds(0, true);
//  AppendChartData(0, vAppend, false);
//  AppendChartData(1, vAppend, true);
//ReplaceChartData(1, vAppend, false, false, true);
//  if (m_pChartDataView != NULL)
//    MAP_SELPNTSD mapSelPntsD = m_pChartDataView->m_mapSelCells;
//  TruncateChart(chartIdx, -10.0, 10.0, true, true);
//  AppendChartData(chartIdx, vData, true); 
//  m_pChartDataView->ShowWindow(SW_MINIMIZE);
//  Sleep(10000);
//  SetDataViewSelPnts(chartIdx, mapSelPntsD);

 UpdateContainerWnds();

}

#endif

void CChartContainer::UpdateDataViewPnts(int chartIdx, size_t dataID, PointD dataPntD, bool bAdd)
{
  bool bRedraw = true;

  if (chartIdx == -1)
  {
    m_mapDataViewPntsD.clear(); 
  }
  else if (m_dataViewChartIdx == chartIdx)
  {
    CChart* chartPtr = GetChart(chartIdx); 
    if (chartPtr != NULL)                   // Chart is not deleted
    {
      if (bAdd)
        m_mapDataViewPntsD.insert(MAP_SELPNTSD::value_type(dataID, dataPntD));
      else
      {
        MAP_SELPNTSD::iterator it = m_mapDataViewPntsD.find(dataID);
        if (it != m_mapDataViewPntsD.end())
          m_mapDataViewPntsD.erase(dataID);
        else
          bRedraw = false;  // Not found the data point
      }
    }
    else
    {
      m_mapDataViewPntsD.clear();
      chartIdx = -1;
    }
  }   // Received right chartIdx
  else
    bRedraw = false;

  if (bRedraw && (NULL != GetVisibleChart(m_dataViewChartIdx)))//IsChartVisible(m_dataViewChartIdx))
    RefreshWnd();
  m_dataViewChartIdx = chartIdx;
}

void CChartContainer::UpdateDataViewPnts(int chartIdx, const MAP_SELPNTSD& mapSelPntsD)
{

  if (chartIdx == m_dataViewChartIdx)
  {
    if (mapSelPntsD.size() > 0)
      m_mapDataViewPntsD = mapSelPntsD;
    else
      m_mapDataViewPntsD.clear();

    if (NULL != GetVisibleChart(chartIdx))
      RefreshWnd();
  }
}

  



